/* -*- indent-tabs-mode: nil; -*- */
#include "config.h"

#include <sys/poll.h>
#include <sys/types.h>
#include <stdio.h>
#include <pthread.h>
#include <stdarg.h>
#include <strings.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>

#include <tcl.h>

#include "common/sgdebug.h"

#include "globus_libc.h"
#include "globus_common.h"
#include "globus_list.h"
#include "globus_xio.h"
#include "globus_xio_types.h"
#include "globus_xio_tcp_driver.h"
#include "globus_xio_gsi.h"

#include "gssapi.h"
#include "globus_openssl.h"
#include "globus_gss_assist.h"
#include "globus_gsi_credential.h"
#include "globus_gsi_callback.h"
#include "globus_gsi_system_config.h"
#include "globus_gsi_proxy.h"
#include "openssl/ssl3.h"

#include "gt_xio_socket.h"

#define XIO_CHANNEL_TYPE_NAME 	"gtxiosock"
#define XIO_CHANNEL_VERSION 	TCL_CHANNEL_VERSION_2

#define NI_MAXHOST	1025
#define SERVICE_NAME	"SERVICE_NAME"
#define X509_CERT_DIR	"X509_CERT_DIR"
#define X509_USER_KEY   "X509_USER_KEY"
#define X509_USER_CERT	"X509_USER_CERT"

#define SEC_THRESHHOLD		0
#define USEC_THRESHHOLD  	10000000

#if SIZEOF_SHORT == 2
typedef short INT_2S;
typedef unsigned short INT_2U;
#else
# error "No 2 byte integer found"
#endif

#if SIZEOF_INT == 4
typedef int INT_4S;
typedef unsigned int INT_4U;
#elif SIZEOF_LONG == 4
typedef long INT_4S;
typedef unsigned long INT_4U;
#else
#   error "No 4 byte integer found"
#endif

#if SIZEOF_LONG == 8
typedef long INT_8S;
typedef unsigned long INT_8U;
#elif SIZEOF_LONG_LONG == 8
typedef long long INT_8S;
typedef unsigned long long INT_8U;
#else
#   error "No 8 byte integer found"
#endif

#if SIZEOF_VOIDP == 4
typedef INT_4U MEM_PTR;
#elif SIZEOF_VOIDP == 8
typedef INT_8U MEM_PTR;
#else
#   error "Unexpected sizeof(void*)"
#endif
#define SIZEOF_MEM_PTR SIZEOF_VOIDP

enum boolean {
  FALSE = 0,
  TRUE
};

/* Main Tcl channel information */
typedef struct mainChannelStruct_ {
  Tcl_Interp*		m_main_interp;
  globus_xio_server_t	m_xio_server;
  globus_xio_attr_t	m_xio_server_attr;
  enum boolean		m_server_mode;
} mainChannelStruct;

/* Client-specific Tcl channel info */
typedef struct channelInfoStruct_ {
  globus_xio_driver_t	m_xio_tcp_driver;
  globus_xio_driver_t	m_xio_gsi_driver;
  globus_xio_stack_t	m_xio_stack;  
  globus_xio_attr_t	m_xio_handle_attr;
} channelInfoStruct;
  
/* Client-specific Tcl channel */
typedef struct channelStruct_ {
  mainChannelStruct*	m_main_channel;
  Tcl_Channel		m_channel;
  char*			m_server_hostname;
  char*			m_service_name;
  char*			m_script;
  globus_xio_handle_t	m_xio_handle;
  channelInfoStruct*	m_channel_info;
  enum boolean		m_delete_on_close;
  enum boolean		m_flip_int_4u;
  enum boolean		m_gsi_auth_enabled;
  globus_reltime_t	m_xio_timeout;
  unsigned int		m_user_timeout;
  globus_xio_gsi_delegation_mode_t	m_delegation_mode;
} channelStruct;

static char HASH_KEY[] = "GTXioAcceptCallbacks";

static int socket_id = 0;	/* Unique socket name identifier */

static enum boolean host_authorization_mode = FALSE;

/*-----------------------------------------------------------------------
 * List of channel functions
 *-----------------------------------------------------------------------*/
static int xio_channel_close_proc( ClientData InstanceData,
                                   Tcl_Interp* Interp );
static int xio_channel_input_proc( ClientData InstanceData,
                                   char* Buffer,
                                   int ToRead,
                                   int* Error );
static int xio_channel_output_proc( ClientData InstanceData,
                                    char* Buffer,
                                    int ToWrite,
                                    int* Error );
static int xio_channel_set_option_proc( ClientData InstanceData,
                                        Tcl_Interp* Interp,
                                        char* OptionName,
                                        char* NewValue );
static void xio_channel_watch_proc( ClientData InstanceData,
                                    int Mask );
static int xio_channel_block_mode_proc( ClientData InstanceData,
                                         int Mode );
static int xio_channel_get_handle_proc( ClientData InstanceData,
                                        int Direction,
                                        ClientData* Handle );
                                  
static int xio_na_proc( );

/*-----------------------------------------------------------------------
 * Forward declaration of local functions
 *-----------------------------------------------------------------------*/
Tcl_ChannelType* get_channel_type( );
char* gen_socket_name( );

static int gt_xio_open( channelStruct* Channel,
                        int Port, char* MyHost, int MyPort );

static int gt_xio_check_error( globus_result_t Return, Tcl_Interp* Interp );

static int gt_xio_check_x509_cert_dir( Tcl_Interp* Interp );

static int gt_xio_check_x509( Tcl_Interp* Interp );

static int gt_xio_activate( Tcl_Interp* Interp ) ;
static int gt_xio_deactivate( Tcl_Interp* Interp );
  
static int gt_xio_init( channelStruct* Channel );

static int gt_xio_create_handle( channelStruct* Channel,
                                 char* MyHost, int MyPort );

static int gt_xio_connect_handle( channelStruct* Channel, int Port );

static int gt_xio_create_server( channelStruct* Channel, int Port );

static void xio_accept( ClientData Data, int Mask );
static void server_close_proc( ClientData Data );
static void delete_proc( ClientData Data, Tcl_Interp* Interp );

static Tcl_Channel get_channel_name( channelStruct* Channel );

static int gt_xio_send_byte_ordering_info( channelStruct* Channel );
static int gt_xio_recv_byte_ordering_info( channelStruct* Channel );
static void gt_xio_flip_int_4u( void* ptr );

static void gt_xio_get_fqdn_host_name( char * Hostname, char * Buffer );

static int gt_xio_get_peer_info( channelStruct* Channel,
                                 char *PeerHostname,
                                 char* PeerIPAddr,
                                 int* PeerPort );

/*-----------------------------------------------------------------------
 * Global function definitions
 *-----------------------------------------------------------------------*/
int
CreateGTXIOSocket( Tcl_Interp* Interp, int ObjC, Tcl_Obj* CONST ObjV[] )
{
  static char function[] = "CreateGTXIOSocket";
  
  char* options[] = {
    "-server",
    "-gsi_auth_enabled",
    "-myaddr",
    "-myport",
    "-host",
    "-timeout",
    "-delegation_mode",
    (char*)NULL
  };

  enum options_type {
    SKT_SERVER = 0,
    SKT_GSI_AUTH_ENABLED,
    SKT_MYADDR,
    SKT_MYPORT,
    SKT_HOST,
    SKT_TIMEOUT,
    SKT_DELEGATION_MODE
  };
  
  enum boolean server_mode = FALSE;
  int	o, indx;
  char* arg = NULL;
  
  char* host = NULL;
  int	host_len;
  int	port = 0;
  
  char* myaddr = NULL;
  int	myaddr_len;
  int	myport = 0;

  char* script = NULL;
  int	script_len;
  
  int	status;
  char* tbuffer = NULL;

  int mytimeout = 0;
  
  enum boolean gsi_auth_enabled = FALSE;

  char* mydelegationmode = NULL;
  int mydelegationmode_len;
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
  /* Parse Tcl options */
  SGDEBUG_LVL_FMT_1( 20, "ObjC: %d", ObjC );
  for ( o = 1; o < ObjC; o++ )
    {
       arg = Tcl_GetString( ObjV[o] );
       if ( *arg != '-' )
        {
          break;
        }
      if ( Tcl_GetIndexFromObj(Interp, ObjV[o], options,
                               "option", TCL_EXACT, &indx) != TCL_OK )
        {
          Tcl_AppendResult( Interp, "Bad options", NULL );
          return TCL_ERROR;
        }
            
      if ( indx == SKT_SERVER )
        {
          server_mode = TRUE;
          SGDEBUG_LVL_FMT_1( 20, "server mode: %d", server_mode );
          o++;
          /* :TODO: Check argument */

          /* Get script name */
          tbuffer = Tcl_GetStringFromObj( ObjV[o], &script_len );
          script = strdup( tbuffer );
          SGDEBUG_LVL_FMT_1( 20, "script: %s", script );
        }
      else if ( indx == SKT_MYADDR )
        {
          o++;
          /* :TODO: Check argument */

          /* Get myaddr name */
          tbuffer = Tcl_GetStringFromObj( ObjV[o], &myaddr_len );
          myaddr = strdup( tbuffer );
          SGDEBUG_LVL_FMT_1( 20, "myaddr: %s", myaddr );
        }
      else if ( indx == SKT_MYPORT )
        {
          o++;
          Tcl_GetInt( Interp, Tcl_GetString( ObjV[o] ), &myport );
          SGDEBUG_LVL_FMT_1( 20, "myport: %d", myport );
        }
      else if ( indx == SKT_HOST )
        {
          SGDEBUG_LVL_FMT( 20, "Setting host authorization mode" );
          host_authorization_mode = TRUE;
        }
      else if ( indx == SKT_GSI_AUTH_ENABLED )
        {
          SGDEBUG_LVL_FMT( 20, "Setting GSI authentication" );
          gsi_auth_enabled = TRUE;
        }
      else if ( indx == SKT_TIMEOUT )
        {
          SGDEBUG_LVL_FMT( 20, "Setting timeout value" );
          o++;
          Tcl_GetInt( Interp, Tcl_GetString( ObjV[o] ), &mytimeout );
          SGDEBUG_LVL_FMT_1( 20, "mytimeout: %d", mytimeout );
        }
      else if ( indx == SKT_DELEGATION_MODE )
        {
          SGDEBUG_LVL_FMT( 20, "Setting delegation mode" );
          o++;
          tbuffer = Tcl_GetStringFromObj( ObjV[o], &mydelegationmode_len );
          mydelegationmode = strdup( tbuffer );
          SGDEBUG_LVL_FMT_1( 20, "mydelegationmode: %s", mydelegationmode );
        }
      else
        {
          Tcl_AppendResult( Interp, "Bad options", NULL );
          return TCL_ERROR;
        }
    }
  
  if ( server_mode == TRUE )
    {
      host = myaddr;
      SGDEBUG_LVL_FMT_1( 20, "host: %s", host );
    }
  else if ( o < ObjC )
    {
      tbuffer= Tcl_GetStringFromObj( ObjV[o++], &host_len );
      host = strdup( tbuffer );
      SGDEBUG_LVL_FMT_1( 20, "host: %s", host );
    }
  else
    {
      return TCL_ERROR;
    }
  
  if ( o == ObjC-1 )
    {
      Tcl_GetInt( Interp, Tcl_GetString( ObjV[o] ), &port );
      SGDEBUG_LVL_FMT_1( 20, "port: %d", port );
    }
  else
    {
      return TCL_ERROR;
    }

  /* Allocate main channel */
  mainChannelStruct* mainChannel =
    (mainChannelStruct*)ckalloc(sizeof(mainChannelStruct));
  SGDEBUG_LVL_FMT_1( 20, "mainChannel: %p", mainChannel );

  mainChannel->m_main_interp = Interp;
  mainChannel->m_server_mode = server_mode;

  /* Allocate XIO-based sub channel info */
  channelInfoStruct* channelInfo =
    (channelInfoStruct*)ckalloc(sizeof(channelInfoStruct));
  SGDEBUG_LVL_FMT_1( 20, "channelInfo: %p", channelInfo );
  
  /* Allocate sub channel */
  channelStruct* subChannel =
    (channelStruct*)ckalloc(sizeof(channelStruct));
  SGDEBUG_LVL_FMT_1( 20, "subChannelPtr: %p", subChannel );
  
  subChannel->m_main_channel = mainChannel;
  subChannel->m_channel_info = channelInfo;

  SGDEBUG_LVL_FMT_1( 20, "Interp: %p", Interp );
  
  if ( server_mode == TRUE )
    {
      SGDEBUG_LVL_FMT( 20, "Server mode" );
      /* Script */
      subChannel->m_script = ckalloc( script_len );
      memset( subChannel->m_script, '\0', script_len );
      strcpy( subChannel->m_script, script );
      SGDEBUG_LVL_FMT_1( 20, "m_script: %s", subChannel->m_script );

      subChannel->m_user_timeout = mytimeout;
      subChannel->m_delete_on_close = FALSE;
      subChannel->m_gsi_auth_enabled = gsi_auth_enabled;
      subChannel->m_service_name = NULL;
      
      if ( subChannel->m_gsi_auth_enabled == TRUE )
        {
          /* Checking for X.509 cert directory */
          status = gt_xio_check_x509_cert_dir( Interp );
          if ( status != TCL_OK )
            {
              return TCL_ERROR;
            }

          /* Checking for X.509 user cert and user key variables */
          status = gt_xio_check_x509( Interp );
          if ( status != TCL_OK )
            {
              return TCL_ERROR;
            }
        }
    }
  else
    {
      SGDEBUG_LVL_FMT( 20, "Client mode" );
      
      subChannel->m_script = NULL;

      /* Server hostname */
      subChannel->m_server_hostname = ckalloc( host_len );
      memset( subChannel->m_server_hostname, '\0', host_len );
      strcpy( subChannel->m_server_hostname, host );
      SGDEBUG_LVL_FMT_1( 20, "m_server_hostname: %s",
                         subChannel->m_server_hostname );

      subChannel->m_user_timeout = mytimeout;
      subChannel->m_delete_on_close = FALSE;
      subChannel->m_gsi_auth_enabled = gsi_auth_enabled;
      
      if ( mydelegationmode )
        {
          if ( ! strcmp(mydelegationmode, "none") )
            {
               subChannel->m_delegation_mode =
                 GLOBUS_XIO_GSI_DELEGATION_MODE_NONE;
            }
          else if ( ! strcmp(mydelegationmode, "limited") )
            {
               subChannel->m_delegation_mode =
                 GLOBUS_XIO_GSI_DELEGATION_MODE_LIMITED;
            }
          else if ( ! strcmp(mydelegationmode, "full") )
            {
               subChannel->m_delegation_mode =
                 GLOBUS_XIO_GSI_DELEGATION_MODE_FULL;
            }
          else
            { 
              Tcl_AppendResult( Interp, "Invalid delegation mode.",
                                NULL );
              return TCL_ERROR;
            }
        }
      else
        {
          subChannel->m_delegation_mode =
            GLOBUS_XIO_GSI_DELEGATION_MODE_NONE;
        }
      
      /* Checking for service name */
      subChannel->m_service_name = Tcl_GetVar( Interp,
                                               SERVICE_NAME,
                                               TCL_GLOBAL_ONLY );

      /* Checking for service name */
      if ( subChannel->m_service_name )
        {
          /* Must be LDAS service */
          SGDEBUG_LVL_FMT_1( 20, "subChannel->m_service_name: %s",
                             subChannel->m_service_name );
#if WORKING      
          Tcl_AppendResult( Interp, 
                            "SERVICE_NAME variable must be set to ldas",
                            NULL );
          return TCL_ERROR;
#endif          
        }
      else
        {
          /* Non LDAS service */
          subChannel->m_service_name = NULL;
        }

      /* Checking for X.509 cert directory */
      if ( host_authorization_mode == TRUE )
        {
          if ( subChannel->m_gsi_auth_enabled == TRUE )
            {
              status = gt_xio_check_x509_cert_dir( Interp );
              if ( status != TCL_OK )
                {
                  return TCL_ERROR;
                }
              /* Checking for X.509 user cert and user key variables */
              status = gt_xio_check_x509( Interp );
              if ( status != TCL_OK )
                {
                  return TCL_ERROR;
                }
            }
        }
    }
  
  /* Open Globus XIO socket */
  status = gt_xio_open( subChannel, port, myaddr, myport );
  if ( status != TCL_OK )
    {
      return TCL_ERROR;
    }

  Tcl_RegisterChannel( Interp, get_channel_name( subChannel ) );
  Tcl_AppendResult( Interp,
                    Tcl_GetChannelName( get_channel_name( subChannel ) ),
                    (char *) NULL);

  if ( ! server_mode && host != NULL )
    {
      SGDEBUG_LVL_FMT( 20, "Releasing local host variable" );
      free( host );
    }
  if ( myaddr != NULL )
    {
      SGDEBUG_LVL_FMT( 20, "Releasing local myaddr variable" );
      free( myaddr );
    }
  if ( script != NULL )
    {
      SGDEBUG_LVL_FMT( 20, "Releasing local script variable" );
      free( script );
    }
  if ( mydelegationmode )
    {
      SGDEBUG_LVL_FMT( 20, "Releasing local mydelegationmode variable" );
      free( mydelegationmode );
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  
  /* Completed successfully */
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 * List of channel functions
 *-----------------------------------------------------------------------*/
static int
xio_channel_close_proc( ClientData InstanceData, Tcl_Interp* Interp )
{
  static char function[] = "xio_channel_close_proc";
  globus_result_t result;
    
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  channelStruct* channelPtr = (channelStruct*)InstanceData;
  SGDEBUG_LVL_FMT_1( 20, "channelPtr: %p", channelPtr );

  mainChannelStruct* mainChannel = channelPtr->m_main_channel;
  SGDEBUG_LVL_FMT_1( 20, "mainChannel: %p", mainChannel );

  channelInfoStruct* channelInfo = channelPtr->m_channel_info;
  SGDEBUG_LVL_FMT_1( 20, "channelInfo: %p", channelInfo );

  if ( channelPtr && Interp )
    {
      if ( mainChannel->m_server_mode == TRUE )
        {
          SGDEBUG_LVL_FMT( 20, "Server mode" );
          
          if ( channelPtr->m_delete_on_close )
            {
              SGDEBUG_LVL_FMT( 20, "m_delete_on_close" );
              
              /*
               * Close client socket for this server and
               * unregister the channel
               */
              
              SGDEBUG_LVL_FMT_1( 20, "channel name: %s",
                                 Tcl_GetChannelName( channelPtr->m_channel ) );

              result = Tcl_UnregisterChannel( Interp,
                                              channelPtr->m_channel );
              SGDEBUG_LVL_FMT_1( 20, "Tcl_UnregisterChannel: result: %d",
                                 result );
              
              result = globus_xio_close( channelPtr->m_xio_handle,
                                         channelInfo->m_xio_handle_attr );
              SGDEBUG_LVL_FMT_1( 20, "globus_xio_close: result: %d",
                                 result );
              
              if ( channelPtr->m_script )
                {
                  SGDEBUG_LVL_FMT( 20, "Deallocate channelPtr->m_script" );
                  free( channelPtr->m_script );
                }

              if ( channelPtr )
                {
                  SGDEBUG_LVL_FMT( 20, "Deallocate channelPtr" );
                  globus_libc_free( channelPtr );
                }
              
            }
          else
            {
              /* Close Globus XIO server handle */
              result = globus_xio_server_close( mainChannel->m_xio_server );
              SGDEBUG_LVL_FMT_1
                ( 20, "globus_xio_server_close: result: %d", result );
              
              result =
                globus_xio_attr_destroy( mainChannel->m_xio_server_attr );
              
              /* Unregister server channel */
              SGDEBUG_LVL_FMT_1( 20, "server channel name: %s",
                                 Tcl_GetChannelName( channelPtr->m_channel ) );
              
              Tcl_UnregisterChannel( Interp, channelPtr->m_channel );
              
              if ( channelPtr->m_script )
                {
                  SGDEBUG_LVL_FMT( 20, "Deallocate channelPtr->m_script" );
                  ckfree( (char*)channelPtr->m_script );
                }
            }
        }
      else
        {
          SGDEBUG_LVL_FMT( 20, "Client mode" );

          SGDEBUG_LVL_FMT_1( 20, "channel name: %s",
                             Tcl_GetChannelName( channelPtr->m_channel ) );
          
          result = Tcl_UnregisterChannel( Interp, channelPtr->m_channel );
          
          SGDEBUG_LVL_FMT_1( 20, "Tcl_UnregisterChannel: result: %d",
                             result );
          
          /* Close Globus XIO handle */
          result = globus_xio_close( channelPtr->m_xio_handle,
                                     channelInfo->m_xio_handle_attr );
          SGDEBUG_LVL_FMT_1( 20, "globus_xio_close: result: %d", result );
          
          if ( channelPtr->m_server_hostname != NULL )
            {
              SGDEBUG_LVL_FMT( 20, "Releasing m_server_hostname storage" );
              ckfree( (char*)channelPtr->m_server_hostname );
            }
        }
      
      if ( ! channelPtr->m_delete_on_close )
        {
          result = globus_xio_attr_destroy(channelInfo->m_xio_handle_attr);
          SGDEBUG_LVL_FMT_1( 20, "Destroyed XIO handle attributes: result: %d",
                             result );
          
          result = globus_xio_stack_destroy(channelInfo->m_xio_stack);
          SGDEBUG_LVL_FMT_1( 20, "Destroyed XIO stack: result: %d", result );
          
          if ( channelPtr->m_gsi_auth_enabled == TRUE )
            {
              result = globus_xio_driver_unload(channelInfo->m_xio_gsi_driver);
              SGDEBUG_LVL_FMT_1( 20, "Finished unloading GSI driver: result: %d",
                                 result );
            }
          
          result = globus_xio_driver_unload(channelInfo->m_xio_tcp_driver);
          SGDEBUG_LVL_FMT_1( 20, "Finished unloading TCP driver: result: %d",
                             result );
          
          /* GT cleanup routine */
          result = gt_xio_deactivate( mainChannel->m_main_interp );
          
          SGDEBUG_LVL_FMT( 20, "Releasing memory resources" );
          
          if ( mainChannel )
            {
              SGDEBUG_LVL_FMT( 20, "Released mainChannel" );
              ckfree( (char*)mainChannel );
            }
          
          if ( channelInfo )
            {
              SGDEBUG_LVL_FMT( 20, "Released channelInfo" );
              ckfree( (char*)channelInfo );
            }
          
          if ( channelPtr )
            {
              SGDEBUG_LVL_FMT( 20, "Released channelPtr" );
              ckfree( (char*)channelPtr );
            }
        }
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  
  /* Completed successfully */
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int 
xio_channel_input_proc( ClientData InstanceData,
                        char* Buffer,
                        int ToRead,
                        int* Error )
{
  static char function[] = "xio_channel_input_proc";
  globus_result_t result;
  globus_size_t nbytes;
  int status;
  globus_size_t wait_for_bytes;
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  channelStruct* channelPtr = (channelStruct*)InstanceData;
  SGDEBUG_LVL_FMT_1( 30, "channelPtr: %p", channelPtr );
  
  mainChannelStruct* mainChannel = channelPtr->m_main_channel;
  SGDEBUG_LVL_FMT_1( 30, "mainChannel: %p", mainChannel );

  SGDEBUG_LVL_FMT_1( 30, "ToRead before globus_xio_read: %d", ToRead );

  /* Read in data length */
  INT_4U dataLength = 0;

#if 1  
  result = globus_xio_read( channelPtr->m_xio_handle,
                            (globus_byte_t*)(&dataLength),
                            sizeof(INT_4U),
                            sizeof(INT_4U),
                            &nbytes,
                            GLOBUS_NULL );
  SGDEBUG_LVL_FMT_1( 30, "globus_xio_read: result: %d", result );
  SGDEBUG_LVL_FMT_1( 30, "nbytes: %d", nbytes );

  if ( result == GLOBUS_SUCCESS )
    {
      /* Check for byte ordering and if necessary flip the bytes */
      if ( channelPtr->m_flip_int_4u == TRUE )
        {
          SGDEBUG_LVL_FMT_1( 30, "Before data length %d", dataLength );
          gt_xio_flip_int_4u( &dataLength );
          SGDEBUG_LVL_FMT_1( 30, "After data length %d", dataLength );
        }
      wait_for_bytes = dataLength;
      
      SGDEBUG_LVL_FMT_1( 30, "wait_for_bytes %d", wait_for_bytes );
  
      /* Read in the actual data */
      result = globus_xio_read( channelPtr->m_xio_handle,
                                (globus_byte_t*)Buffer,
                                wait_for_bytes,
                                wait_for_bytes,
                                (globus_size_t*)&ToRead,
                                GLOBUS_NULL );
      SGDEBUG_LVL_FMT_1( 30, "globus_xio_read: result: %d", result );
    }
  else
    {
      ToRead = 0;
    }
  
  SGDEBUG_LVL_FMT_1( 30, "ToRead after globus_xio_read: %d", ToRead );
  
#else
  
  int bytes_read;
  
  ToRead = 0;
  while( 1 )
    {
      result = globus_xio_read( channelPtr->m_xio_handle,
                                (globus_byte_t*)(&dataLength),
                                sizeof(INT_4U),
                                sizeof(INT_4U),
                                &nbytes,
                                GLOBUS_NULL );
      SGDEBUG_LVL_FMT_1( 30, "globus_xio_read: result: %d", result );
      SGDEBUG_LVL_FMT_1( 30, "nbytes: %d", nbytes );

      if ( result == GLOBUS_SUCCESS )
        {
          fprintf( stderr, "DEBUG: nbytes: %d (%d)\n", nbytes, dataLength );
          /* Check for byte ordering and if necessary flip the bytes */
          if ( channelPtr->m_flip_int_4u == TRUE )
            {
              SGDEBUG_LVL_FMT_1( 30, "Before data length %d", dataLength );
              gt_xio_flip_int_4u( &dataLength );
              SGDEBUG_LVL_FMT_1( 30, "After data length %d", dataLength );
            }
          wait_for_bytes = dataLength;
          
          SGDEBUG_LVL_FMT_1( 30, "wait_for_bytes %d", wait_for_bytes );
          
          /* Read in the actual data */
          result = globus_xio_read( channelPtr->m_xio_handle,
                                    &(Buffer[ToRead]),
                                    wait_for_bytes,
                                    wait_for_bytes,
                                    &bytes_read,
                                    GLOBUS_NULL );
          ToRead += bytes_read;
          fprintf( stderr, "DEBUG: ToRead after globus_xio_read(1): %d\n",
                   ToRead );
          /* Schedule another interupt when data is available */
          {
            channelInfoStruct* channelInfo = channelPtr->m_channel_info;
            globus_xio_system_handle_t system_handle;
            struct pollfd  fds[1];

            result = globus_xio_handle_cntl( channelPtr->m_xio_handle,
                                             channelInfo->m_xio_tcp_driver,
                                             GLOBUS_XIO_TCP_GET_HANDLE,
                                             &system_handle );
            fds[0].fd = (int)system_handle;
            fds[0].events = POLLIN;
            if ( poll( fds, sizeof( fds ) / sizeof( *fds ), 100 ) <= 0 )
              {
                /* Nothing left to read */
                fprintf( stderr, "DEBUG: globus_xio_read: Done reading\n");
                break;
              }
          }
          fprintf( stderr, "DEBUG: globus_xio_read: Going to read some more\n");
          SGDEBUG_LVL_FMT_1( 30, "globus_xio_read: result: %d", result );
        }
      else
        {
          break;
        }
    }
  
  fprintf( stderr, "DEBUG: ToRead after globus_xio_read: %d\n", ToRead );
  SGDEBUG_LVL_FMT_1( 30, "ToRead after globus_xio_read: %d", ToRead );
#endif
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  
  /* Completed successfully */
  return ToRead;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int 
xio_channel_output_proc( ClientData InstanceData,
                         char* Buffer,
                         int ToWrite,
                         int* Error )
{
  static char function[] = "xio_channel_output_proc";
  globus_result_t result;
  int status;
  int nbytes;
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  channelStruct* channelPtr = (channelStruct*)InstanceData;
  SGDEBUG_LVL_FMT_1( 30, "channelPtr: %p", channelPtr );

  mainChannelStruct* mainChannel = channelPtr->m_main_channel;
  SGDEBUG_LVL_FMT_1( 30, "mainChannel: %p", mainChannel );
  
  SGDEBUG_LVL_FMT_1( 30, "channel buffer size: %d",
                     Tcl_GetChannelBufferSize( channelPtr->m_channel ) );

  SGDEBUG_LVL_FMT_1( 30, "ToWrite: %d", ToWrite );
  
  if ( ToWrite != 0 )
    {
      /* Send out data length information */
      INT_4U buf;
      
      buf = ToWrite;
      SGDEBUG_LVL_FMT_1( 30, "buf: %d", buf );
      
      result = globus_xio_write( channelPtr->m_xio_handle,
                                 (globus_byte_t*)(&buf),
                                 sizeof(INT_4U),
                                 sizeof(INT_4U),
                                 (globus_size_t*)&nbytes,
                                 GLOBUS_NULL );
      SGDEBUG_LVL_FMT_1( 30, "globus_xio_write: result: %d", result );
      SGDEBUG_LVL_FMT_1( 30, "nbytes: %d", nbytes );

      /* Send out the actual data */
      result = globus_xio_write( channelPtr->m_xio_handle,
                                 (globus_byte_t*)Buffer,
                                 ToWrite,
                                 ToWrite,
                                 (globus_size_t*)&nbytes,
                                 GLOBUS_NULL );
      SGDEBUG_LVL_FMT_1( 30, "globus_xio_write: result: %d", result );
      SGDEBUG_LVL_FMT_1( 30, "nbytes: %d", nbytes );
#if 0
      sleep(1);
#endif      
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  
  /* Completed successfully */
  return ToWrite;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
xio_channel_set_option_proc( ClientData InstanceData,
                             Tcl_Interp* Interp,
                             char* OptionName,
                             char* NewValue )
{
  static char function[] = "xio_channel_set_option_proc";
  globus_result_t result;
  int status = TCL_OK;
  int secs;
  int usecs;
  
  char* options =
    "-linger"
    "-timeout"
    "-gsi-auth-enabled";

  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  channelStruct* channelPtr = (channelStruct*)InstanceData;
  SGDEBUG_LVL_FMT_1( 30, "channelPtr: %p", channelPtr );

  mainChannelStruct* mainChannel = channelPtr->m_main_channel;
  SGDEBUG_LVL_FMT_1( 30, "mainChannel: %p", mainChannel );
  
  channelInfoStruct* channelInfo = channelPtr->m_channel_info;
  SGDEBUG_LVL_FMT_1( 30, "channelInfo: %p", channelInfo );

  SGDEBUG_LVL_FMT_1( 20, "OptionName: %s", OptionName );
  SGDEBUG_LVL_FMT_1( 20, "NewValue: %s", NewValue );
  
  if ( strcmp( OptionName, "-linger" ) == 0 )
    {
      result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                     channelInfo->m_xio_tcp_driver,
                                     GLOBUS_XIO_TCP_SET_LINGER,
                                     GLOBUS_TRUE,
                                     atoi(NewValue) );
      SGDEBUG_LVL_FMT_1( 20, "GLOBUS_XIO_TCP_SET_LINGER: result: %d",
                         result );
      
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
        }
    }
  
  else if ( strcmp( OptionName, "-timeout" ) == 0 )
    {
      /* Convert the value from milliseconds to microseconds */
      int timeout = atoi(NewValue) * 1000;
      
      GlobusTimeReltimeSet(channelPtr->m_xio_timeout, SEC_THRESHHOLD, timeout);
      GlobusTimeReltimeGet(channelPtr->m_xio_timeout, secs, usecs);
      SGDEBUG_LVL_FMT_2( 20, "new timeout values: secs: %d, usecs: %d ",
                         secs, usecs );

      result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                     NULL,
                                     GLOBUS_XIO_ATTR_SET_TIMEOUT_ALL,
                                     NULL,
                                     &(channelPtr->m_xio_timeout),
                                     NULL );
      SGDEBUG_LVL_FMT_1( 20, "GLOBUS_XIO_ATTR_SET_TIMEOUT_ALL: result: %d",
                         result );
      
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
        }
    }
  
  else if ( strcmp( OptionName, "-gsi-auth-enabled" ) == 0 )
    {
      channelPtr->m_gsi_auth_enabled = TRUE;
    }
  else
    {
      status = Tcl_BadChannelOption( Interp, OptionName, options );
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  
  /* Completed successfully */
  return status;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
xio_channel_get_option_proc( ClientData InstanceData,
                             Tcl_Interp* Interp,
                             char* OptionName,
                             Tcl_DString* Result )
{
  static char function[] = "xio_channel_get_option_proc";
  globus_result_t result;
  int status = TCL_OK;
  char* options =
    "-peername"
    "-readable"
    "-x509_subj_name";
  char* contact = NULL;
  char peer_ipaddr[256];
  char peer_hostname[256];
  int peer_port;

  char* peerIP = NULL;
  char* peerHostname = NULL;
  char* peerPort = NULL;
  char  buffer[512];
  gss_name_t peer_name;
  gss_buffer_desc peer_name_buffer;
  OM_uint32 major_status = GSS_S_COMPLETE;
  OM_uint32 minor_status;
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  SGDEBUG_LVL_FMT_1( 20, "OptionName: %s", OptionName );

  channelStruct* channelPtr = (channelStruct*)InstanceData;
  SGDEBUG_LVL_FMT_1( 30, "channelPtr: %p", channelPtr );

  mainChannelStruct* mainChannel = channelPtr->m_main_channel;
  SGDEBUG_LVL_FMT_1( 30, "mainChannel: %p", mainChannel );
  
  channelInfoStruct* channelInfo = channelPtr->m_channel_info;
  SGDEBUG_LVL_FMT_1( 30, "channelInfo: %p", channelInfo );
  
  if ( strcmp( OptionName, "-peername" ) == 0 )
    {
      /*========================================================*/
      /* Get peer information					*/
      /*========================================================*/
      status = gt_xio_get_peer_info( channelPtr, peer_hostname,
                                     peer_ipaddr, &peer_port );
      if ( status != TCL_OK )
        {
          /* Inform user for error */
        }
          
      SGDEBUG_LVL_FMT_1( 20, "peer_hostname: %s", peer_hostname );
      SGDEBUG_LVL_FMT_1( 20, "peer_ipaddr: %s", peer_ipaddr );
      SGDEBUG_LVL_FMT_1( 20, "peer_port: %d", peer_port );
      sprintf( buffer, "%d", peer_port );
      
      Tcl_DStringAppendElement( Result, peer_ipaddr );
      Tcl_DStringAppendElement( Result, peer_hostname );
      Tcl_DStringAppendElement( Result, buffer );

    }

  else if ( strcmp( OptionName, "-readable" ) == 0 )
    {
      globus_xio_system_handle_t system_handle;
      struct pollfd fds[1];
      
      result = globus_xio_handle_cntl( channelPtr->m_xio_handle,
                                       channelInfo->m_xio_tcp_driver,
                                       GLOBUS_XIO_TCP_GET_HANDLE,
                                       &system_handle );
      fds[0].fd = (int)system_handle;
      fds[0].events = POLLIN;
      if ( poll( fds, sizeof( fds ) / sizeof( *fds ), 0 ) <= 0 )
        {
          /* Nothing left to read */
          Tcl_DStringAppendElement( Result, "0" );
        }
      else
        {
          Tcl_DStringAppendElement( Result, "1" );
        }
    }
  
  else if ( strcmp( OptionName, "-x509_subj_name" ) == 0 )
    {
      if ( channelPtr->m_gsi_auth_enabled == TRUE )
        {
          result = globus_xio_handle_cntl( channelPtr->m_xio_handle,
                                           channelInfo->m_xio_gsi_driver,
                                           GLOBUS_XIO_GSI_GET_PEER_NAME,
                                           &peer_name );
          SGDEBUG_LVL_FMT_1( 20, "globus_xio_handle_cntl: result: %d",
                             result );
          if ( result != GLOBUS_SUCCESS )
            {
              status = gt_xio_check_error( result,
                                           mainChannel->m_main_interp );
              if ( status != TCL_OK )
                {
                  return;
                }
            }
          
          major_status = gss_display_name( &minor_status,
                                           peer_name,
                                           &peer_name_buffer,
                                           (gss_OID*) NULL );

          if ( major_status != GSS_S_COMPLETE )
            {
              /* Inform user for error */
              Tcl_DStringAppendElement( Result,
                                        "Can't retrieve peer X.509 subject name" );
              return;
            }
          
          sprintf( buffer, "\"%.*s\"",
                   (int)peer_name_buffer.length,
                   (char*)peer_name_buffer.value );
          SGDEBUG_LVL_FMT_1( 20, "X.509 peer subject name: %s", buffer );
          
          gss_release_buffer( &minor_status, &peer_name_buffer );
          
          Tcl_DStringAppendElement( Result, buffer );
        }
      else
        {
          Tcl_DStringAppendElement
            ( Result,
              "Bad option: -x509_subj_name can't be used with non-GSI authentication method" );
        }
    }
  
  else
    {
      status = Tcl_BadChannelOption( Interp, OptionName, options );
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );

  /* Completed successfully */
  return status;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static void
xio_channel_watch_proc( ClientData InstanceData,
                        int Mask )
{
  static char function[] = "xio_channel_watch_proc";
  
  globus_xio_system_handle_t system_handle;
  globus_result_t result;
  char* server_cs = NULL;
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  channelStruct* channelPtr = (channelStruct*)InstanceData;
  SGDEBUG_LVL_FMT_1( 30, "channelPtr: %p", channelPtr );

  channelInfoStruct* channelInfo = channelPtr->m_channel_info;
  SGDEBUG_LVL_FMT_1( 30, "channelInfo: %p", channelInfo );
  
  SGDEBUG_LVL_FMT_1( 20, "Mask: %d", Mask );

  SGDEBUG_LVL_FMT_1( 20, "m_channel: %s",
                     Tcl_GetChannelName( channelPtr->m_channel ) );
  
  SGDEBUG_LVL_FMT_1( 20, "channelPtr->m_xio_handle: %p",
                     channelPtr->m_xio_handle );
  
  result = globus_xio_handle_cntl( channelPtr->m_xio_handle,
                                   channelInfo->m_xio_tcp_driver,
                                   GLOBUS_XIO_TCP_GET_HANDLE,
                                   &system_handle );
  SGDEBUG_LVL_FMT_1( 20, "globus_xio_handle_cntl: result: %d", result );
  SGDEBUG_LVL_FMT_1( 20, "system_handle: %d", system_handle );

  if ( Mask )
    {
      SGDEBUG_LVL_FMT( 20, "Prior to calling Tcl_CreateFileHandler" );
      Tcl_CreateFileHandler( (int)system_handle,
                             Mask,
                             (Tcl_FileProc *) Tcl_NotifyChannel,
                             (ClientData)channelPtr->m_channel );
    }
  else
    {
      SGDEBUG_LVL_FMT( 20, "Prior to calling Tcl_DeleteFileHandler" );
      Tcl_DeleteFileHandler( system_handle );
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
xio_channel_get_handle_proc( ClientData InstanceData,
                                        int Direction,
                                        ClientData* Handle )
{
  static char function[] = "xio_channel_get_handle_proc";
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
xio_channel_block_mode_proc( ClientData InstanceData,
                             int Mode )
{
  static char function[] = "xio_channel_block_mode_proc";
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  SGDEBUG_LVL_FMT_1( 20, "Mode %d", Mode );
  
  if ( Mode == TCL_MODE_BLOCKING )
    {
      SGDEBUG_LVL_FMT( 20, "TCL_MODE_BLOCKING" );
    }
  else
    {
      SGDEBUG_LVL_FMT( 20, "TCL_MODE_NONBLOCKING" );
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );

  return 0;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
xio_na_proc( )
{
  return EINVAL;
}

/*-----------------------------------------------------------------------
 * Local function declarations
 *-----------------------------------------------------------------------*/

Tcl_ChannelType*
get_channel_type( )
{
  static char function[] = "get_channel_type";
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
  static Tcl_ChannelType data = {
    XIO_CHANNEL_TYPE_NAME,			/* Type name */
    XIO_CHANNEL_VERSION,			/* Version number */
    xio_channel_close_proc,			/* Close proc */
    xio_channel_input_proc,			/* Tcl_DriverInputProc */
    xio_channel_output_proc,			/* Tcl_DriverOutputProc */
    (Tcl_DriverSeekProc*)xio_na_proc,		/* Tcl_DriverSeekProc */
    xio_channel_set_option_proc,		/* Tcl_DriverSetOptionProc */
    xio_channel_get_option_proc,		/* Tcl_DriverGetOptionProc */
    xio_channel_watch_proc,			/* Tcl_DriverWatchProc */
    xio_channel_get_handle_proc,		/* Tcl_DriverGetHandleProc */
    (Tcl_DriverClose2Proc*)xio_na_proc,		/* Tcl_DriverClose2Proc */
    xio_channel_block_mode_proc,		/* Tcl_DriverBlockModeProc */
    (Tcl_DriverFlushProc*)xio_na_proc,		/* Tcl_DriverFlushProc */
    (Tcl_DriverHandlerProc*)xio_na_proc,	/* Tcl_DriverHandlerProc */
#if HAVE_TCL_DRIVER_WIDE_SEEK_PROC    
    (Tcl_DriverWideSeekProc*)xio_na_proc	/* Tcl_DriverWideSeekProc */
#endif /* HAVE_TCL_DRIVER_WIDE_SEEK_PROC */    
  };

  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  
  return &data;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
char*
gen_socket_name( )
{
  static char Buffer[64];
  static char function[] = "gen_socket_name";
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  sprintf( Buffer, "%s%d", XIO_CHANNEL_TYPE_NAME, socket_id++ );
  SGDEBUG_LVL_FMT_1( 20, "Buffer: %s", Buffer );
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  return (Buffer);
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static Tcl_Channel
get_channel_name( channelStruct* Channel )
{
  return ( Channel->m_channel );
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
gt_xio_check_error( globus_result_t Result, Tcl_Interp* Interp )
{
  char* err;
  static char function[] = "gt_xio_check_error";

  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
  if ( Result != GLOBUS_SUCCESS )
    {
      err = globus_object_printable_to_string( globus_error_get( Result ) );
      Tcl_SetObjResult( Interp, Tcl_NewStringObj(err, -1) );
      globus_free( err );
      SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
      return TCL_ERROR;
    }
  else
    {
      SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
      return TCL_OK;
    }
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
gt_xio_activate( Tcl_Interp* Interp )
{
  static char function[] = "gt_xio_activate";
  globus_result_t result;
  int status = TCL_OK;
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  result = globus_module_activate(GLOBUS_XIO_MODULE);
  
  SGDEBUG_LVL_FMT_1( 20, "Activating XIO module: result: %d", result );

  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, Interp );
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  
  return status;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
gt_xio_deactivate( Tcl_Interp* Interp )
{
  static char function[] = "gt_xio_deactivate";
  globus_result_t result;
  int status = TCL_OK;
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  result = globus_module_deactivate(GLOBUS_XIO_MODULE);
  
  SGDEBUG_LVL_FMT_1( 20, "Deactivating XIO module: result: %d", result );

  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, Interp );
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  
  return status;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
gt_xio_send_byte_ordering_info( channelStruct* Channel )
{
  static char function[] = "gt_xio_send_byte_ordering_info";

  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  globus_result_t result;
  INT_2U byte_ordering = 1;
  globus_size_t nbytes;
  int status = TCL_OK;

  mainChannelStruct*  mainChannel = Channel->m_main_channel;
  SGDEBUG_LVL_FMT_1( 30, "mainChannel: %p", mainChannel );

  /* Sends out byte ordering information to the receiving end */
  result = globus_xio_write( Channel->m_xio_handle,
                             (globus_byte_t*)(&byte_ordering),
                             sizeof(INT_2U),
                             sizeof(INT_2U),
                             &nbytes,
                             GLOBUS_NULL );
  SGDEBUG_LVL_FMT_1( 20, "globus_xio_write: result: %d", result );
  SGDEBUG_LVL_FMT_1( 20, "nbytes: %d", nbytes );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );

  return status;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
gt_xio_recv_byte_ordering_info( channelStruct* Channel )
{
  static char function[] = "gt_xio_recv_byte_ordering_info";

  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
   /* Get byte ordering information from sending end */
  INT_2U* buf = globus_libc_malloc( sizeof(INT_2U) );
  globus_size_t nbytes;
  globus_result_t result;
  int status = TCL_OK;
  
  mainChannelStruct*  mainChannel = Channel->m_main_channel;
  SGDEBUG_LVL_FMT_1( 30, "mainChannel: %p", mainChannel );

  result = globus_xio_read( Channel->m_xio_handle,
                            (globus_byte_t*)(buf),
                            sizeof(INT_2U),
                            sizeof(INT_2U),
                            &nbytes,
                            GLOBUS_NULL );
  SGDEBUG_LVL_FMT_1( 20, "globus_xio_read: result: %d", result );
  SGDEBUG_LVL_FMT_1( 20, "nbytes: %d", nbytes );
  if ( result != GLOBUS_SUCCESS )
    {
      globus_libc_free( buf );
      /* status = gt_xio_check_error( result, mainChannel->m_main_interp ); */
      /* if ( status != TCL_OK ) */
      /* { */
      Tcl_AppendResult( mainChannel->m_main_interp, 
                        "Your authentication doesn't match with the one the server is using",
                        NULL );
      return TCL_ERROR;
      /* } */
    }
  
  SGDEBUG_LVL_FMT_1( 20, "buf: %u", *buf );

  if ( *buf != (INT_2U)1 )
    {
      /* Set the flip_byte flag */
      SGDEBUG_LVL_FMT( 20, "Set the flip_byte flag for this channel" );
      Channel->m_flip_int_4u = TRUE;
    }
  else
    {
      Channel->m_flip_int_4u = FALSE;
    }
  
  globus_libc_free( buf );
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  
  return status;  
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static void
gt_xio_flip_int_4u( void* ptr )
{
  static char function[] = "gt_xio_flip_int_4u";

  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  SGDEBUG_LVL_FMT_1( 20, "Before ptr: %d", *(INT_4U*)ptr );

  char* d = ptr;
  char tmp;
  tmp = *d;
  *d = *( d + 3 );
  *( d + 3 ) = tmp;
  tmp = *( d + 1 );
  *( d + 1 ) = *( d + 2 );
  *( d + 2 ) = tmp;
  d += 4;

  SGDEBUG_LVL_FMT_1( 20, "After ptr: %d", *(INT_4U*)ptr );
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/

static void
gt_xio_get_fqdn_host_name( char * Hostname, char * Buffer )
{
  globus_addrinfo_t	hints;
  globus_addrinfo_t *	addrinfo;
  char			realhostname[NI_MAXHOST];
  globus_result_t	result;
  
  static char function[] = "gt_xio_get_fqdn_host_name";
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
  memset( realhostname, 0, NI_MAXHOST);
  
  memset(&hints, 0, sizeof(globus_addrinfo_t));
  hints.ai_family = PF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_protocol = 0;

#if WORKING  
  /* If hostname is an ip address, do a non-canonname getaddrinfo
   * to get the sockaddr, then getnameinfo to get the hostname
   * from that addr */ 
  hints.ai_flags = GLOBUS_AI_NUMERICHOST;
  result = globus_libc_getaddrinfo(Hostname, NULL, &hints, &addrinfo);
  /* if this succeeds then the hostname must be numeric */
  if ( result == GLOBUS_SUCCESS )
    {
      /* use GLOBUS_NI_NAMEREQD to fail if address can't be looked up? 
       * if not, realhostname will just be the same ip address 
       * we pass in */
      result =
        globus_libc_getnameinfo
        ( (const globus_sockaddr_t *) addrinfo->ai_addr,
          &realhostname[0],
          sizeof(realhostname),
          NULL,
          0,
          0 );
      
      globus_libc_freeaddrinfo(addrinfo);
      
      SGDEBUG_LVL_FMT_1( 20, "realhostname: %s", realhostname );
    }
#endif
  
  /* just do a getaddrinfo lookup of the hostname */
  hints.ai_flags = GLOBUS_AI_CANONNAME;
  result = globus_libc_getaddrinfo(Hostname, NULL, &hints, &addrinfo);
  strncpy(&realhostname[0], addrinfo->ai_canonname, 
          sizeof(realhostname));
  realhostname[sizeof(realhostname) - 1] = '\0';
  
  globus_libc_freeaddrinfo(addrinfo);
  
  SGDEBUG_LVL_FMT_1( 20, "realhostname: %s", realhostname );

  sprintf( Buffer, "%s", realhostname );
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
gt_xio_check_x509_cert_dir( Tcl_Interp* Interp )
{
  static char function[] = "gt_xio_check_x509_cert_dir";
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  const char * x509_cert_dir = Tcl_GetVar( Interp, X509_CERT_DIR,
                                           TCL_GLOBAL_ONLY );

  if ( x509_cert_dir == NULL )
    {
      if ( host_authorization_mode == TRUE )
        {
          /* Checking /etc/grid-security/certificates ... */
          x509_cert_dir = "/etc/grid-security/certificates";
        }
      
      if ( access(x509_cert_dir, F_OK) != 0 )
        {
          SGDEBUG_LVL_FMT( 20, "x509_cert_dir can't be accessed" );
          Tcl_AppendResult( Interp, x509_cert_dir, " can't be accessed",
                            NULL );
          return TCL_ERROR;
        }
    }
  else
    {
      if ( access(x509_cert_dir, F_OK) != 0 )
        {
          SGDEBUG_LVL_FMT( 20, "x509_cert_dir " );
          Tcl_AppendResult( Interp, x509_cert_dir, " does not exist",
                            NULL );
          return TCL_ERROR;
        }
      else
        {
          if( setenv("X509_CERT_DIR", x509_cert_dir, 1) != 0 )
            {
              Tcl_AppendResult( Interp, 
                                "Can't set X509_CERT_DIR env variable",
                                NULL );
              return TCL_ERROR;
            }
        }
    }

  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
gt_xio_check_x509( Tcl_Interp* Interp )
{
  static char function[] = "gt_xio_check_x509";
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
  /* Check for X509_USER_KEY LDAS resource variable */
  const char * x509_user_key = Tcl_GetVar( Interp, X509_USER_KEY,
                                           TCL_GLOBAL_ONLY );

  /* Check for X509_USER_CERT LDAS resource variable */
  const char * x509_user_cert = Tcl_GetVar( Interp, X509_USER_CERT,
                                            TCL_GLOBAL_ONLY );
  
  if ( x509_user_key == NULL )
    {
#if WORKING      
      Tcl_AppendResult( Interp, "Missing X509_USER_KEY resource variable",
                        NULL );
      return TCL_ERROR;
#endif      
    }
  else if ( x509_user_cert == NULL )
    {
#if WORKING      
      Tcl_AppendResult( Interp, "Missing X509_USER_CERT resource variable",
                        NULL );
      return TCL_ERROR;
#endif      
    }
  else 
    {
      if ( x509_user_key )
        {
          SGDEBUG_LVL_FMT_1( 20, "x509_user_key: %s", x509_user_key );
          if ( access(x509_user_key, F_OK) != 0 )
            {
              Tcl_AppendResult( Interp, 
                                x509_user_key, " doesn't exist",
                                NULL );
              return TCL_ERROR;
            }
          else
            {
              if ( setenv("X509_USER_KEY", x509_user_key, 1) != 0 )
                {
                  Tcl_AppendResult( Interp, 
                                    "Can't set X509_USER_KEY env variable",
                                    NULL );
                  return TCL_ERROR;
                }
            }
        }
      else
        {
          Tcl_AppendResult( Interp, 
                            "X509_USER_KEY Tcl variable is not defined",
                            NULL );
          return TCL_ERROR;
          
        }
      
      if ( x509_user_cert )
        {
          SGDEBUG_LVL_FMT_1( 20, "x509_user_cert: %s", x509_user_cert );
          if ( access(x509_user_cert, F_OK) != 0 )
            {
              Tcl_AppendResult( Interp, 
                                x509_user_cert, " doesn't exist",
                                NULL );
              return TCL_ERROR;
            }
          else
            {
              if ( setenv("X509_USER_CERT", x509_user_cert, 1) != 0 )
                {
                  Tcl_AppendResult( Interp, 
                                    "Can't set X509_USER_CERT env variable",
                                    NULL );
                  return TCL_ERROR;
                }
            }
        }
      else
        {
          Tcl_AppendResult( Interp, 
                            "X509_USER_CERT Tcl variable is not defined",
                            NULL );
          return TCL_ERROR;
        }
    }

  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
gt_xio_open( channelStruct* Channel, int Port, char* MyHost, int MyPort )
{
  static char function[] = "gt_xio_open";
  globus_result_t 		result;
  int				status = TCL_OK;
  globus_xio_system_handle_t	fd;
  
  int 				created;
  Tcl_Interp*			interp;
  Tcl_HashTable*       		hTblPtr;
  Tcl_HashEntry*		hPtr;
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  mainChannelStruct*  mainChannel = Channel->m_main_channel;
  SGDEBUG_LVL_FMT_1( 30, "mainChannel: %p", mainChannel );
  
  channelInfoStruct*  channelInfo = Channel->m_channel_info;
  SGDEBUG_LVL_FMT_1( 30, "channelInfo: %p", channelInfo );
  
  status = gt_xio_activate( mainChannel->m_main_interp );
  if ( status != TCL_OK )
    {
      return TCL_ERROR;
    }
  
  /* Server mode */
  if ( mainChannel->m_server_mode == TRUE )
    {
      SGDEBUG_LVL_FMT( 20, "Server mode" );
      
      /* Get GT XIO Server started */
      status = gt_xio_create_server( Channel, Port );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
      
      /*
       * Set up callback mechanism for accepting connections from new client
       */
      result = globus_xio_server_cntl( mainChannel->m_xio_server,
                                       channelInfo->m_xio_tcp_driver,
                                       GLOBUS_XIO_TCP_GET_HANDLE,
                                       &fd );
      SGDEBUG_LVL_FMT_1( 20, "globus_xio_server_cntl: result: %d", result );
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
          if ( status != TCL_OK )
            {
              return status;
            }
        }
      SGDEBUG_LVL_FMT_1( 20, "globus_xio_server_cntl: fd: %d", fd );
      
      if ( result == GLOBUS_SUCCESS )
        {
          SGDEBUG_LVL_FMT_1( 20, "m_script: %s", Channel->m_script );
          SGDEBUG_LVL_FMT_1( 20, "m_main_interp: %p",
                             mainChannel->m_main_interp );
          
          Tcl_CreateFileHandler( (int)fd,
                                 TCL_READABLE,
                                 xio_accept,
                                 (ClientData)Channel );
          
          /* Create server channel */
          Channel->m_channel =
            Tcl_CreateChannel( get_channel_type( ),
                               gen_socket_name( ),
                               (ClientData)Channel,
                               (TCL_READABLE | TCL_WRITABLE) );
          
          if ( Channel->m_channel )
            {
              SGDEBUG_LVL_FMT( 20, "Create server channel" );

              /*
               * Register with the interpreter to let us know when the
               * interpreter is deleted (by having the callback set the
               * m_main_interp field to NULL). This is to avoid trying to
               * eval the script in a deleted interpreter.
               */
              
              interp = mainChannel->m_main_interp;
              hTblPtr =
                (Tcl_HashTable*)Tcl_GetAssocData( interp, HASH_KEY, NULL );
              
              if ( hTblPtr == (Tcl_HashTable*)NULL )
                {
                  hTblPtr =
                    (Tcl_HashTable*)ckalloc((unsigned)sizeof(Tcl_HashTable));
                  Tcl_InitHashTable( hTblPtr, TCL_ONE_WORD_KEYS );
                  
                  Tcl_SetAssocData( interp,
                                    HASH_KEY,
                                    delete_proc,
                                    (ClientData)hTblPtr );
                  
                }
              
              hPtr = Tcl_CreateHashEntry( hTblPtr,
                                          (char*)Channel,
                                          &created );
              if ( !created )
                {
                  panic( "GT XIO Server Cleanup: damaged accept record table" );
                  return TCL_ERROR;
                }
              Tcl_SetHashValue( hPtr, (ClientData)Channel );
              
              /*
               * Register a close callback. This callback will inform the
               * interpreter (if it still exists) that this channel does not
               * need to be informed when the interpreter is deleted
               */
              Tcl_CreateCloseHandler( Channel->m_channel,
                                      server_close_proc,
                                      (ClientData)Channel );
            }
        }
    }
  else	/* Client mode */
    {
      SGDEBUG_LVL_FMT( 20, "Client mode" );
      SGDEBUG_LVL_FMT_1( 20, "channel: %p", Channel );
      
      /*
       * Creates Globus XIO handle at the specified port
       * PURE GLOBUS calls
       */
      status = gt_xio_create_handle( Channel,
                                     MyHost, MyPort );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }

      /*
       * Connects the given handle to a server located at the specified
       * address and port
       * PURE GLOBUS calls
       */
      status = gt_xio_connect_handle( Channel, Port );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
      
      /* Actually create client channel */
      Channel->m_channel =
        Tcl_CreateChannel( get_channel_type( ),
                           gen_socket_name( ),
                           (ClientData)Channel,
                           (TCL_READABLE | TCL_WRITABLE) );

    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  
  /* Completed successfully */
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
gt_xio_init( channelStruct* Channel )
{
  static char function[] = "gt_xio_init";
  globus_result_t result;
  int status = TCL_OK;
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
  mainChannelStruct*  mainChannel = Channel->m_main_channel;
  SGDEBUG_LVL_FMT_1( 30, "mainChannel: %p", mainChannel );
  
  channelInfoStruct*  channelInfo = Channel->m_channel_info;
  SGDEBUG_LVL_FMT_1( 30, "channelInfo: %p", channelInfo );
  
  result = globus_xio_driver_load( "tcp", &(channelInfo->m_xio_tcp_driver) );
  SGDEBUG_LVL_FMT_1( 20, "Loading TCP driver: result: %d", result );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }

  if ( Channel->m_gsi_auth_enabled == TRUE )
    {
      result = globus_xio_driver_load( "gsi",
                                       &(channelInfo->m_xio_gsi_driver) );
      SGDEBUG_LVL_FMT_1( 20, "Loading GSI driver: result: %d", result );
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
          if ( status != TCL_OK )
            {
              return TCL_ERROR;
            }
        }
    }

  result = globus_xio_stack_init( &(channelInfo->m_xio_stack), NULL );
  SGDEBUG_LVL_FMT_1( 20, "Creating stack: result: %d", result );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }

  result = globus_xio_stack_push_driver( channelInfo->m_xio_stack,
                                         channelInfo->m_xio_tcp_driver );
  SGDEBUG_LVL_FMT_1( 20, "Pushing TCP driver onto stack: result: %d",
                     result );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }

  if ( Channel->m_gsi_auth_enabled == TRUE )
    {
      result = globus_xio_stack_push_driver( channelInfo->m_xio_stack,
                                             channelInfo->m_xio_gsi_driver );
      SGDEBUG_LVL_FMT_1( 20, "Pushing GSI driver onto stack: result: %d",
                         result );
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
          if ( status != TCL_OK )
            {
              return TCL_ERROR;
            }
        }
    }

  result = globus_xio_attr_init( &(channelInfo->m_xio_handle_attr) );
  SGDEBUG_LVL_FMT_1( 20, "Initializing handle attributes: result: %d",
                     result );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );

  /* Completed successfully */
  return TCL_OK;  
}

/*-----------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------*/
static int
gt_xio_create_handle( channelStruct* Channel, char* MyHost, int MyPort )
{
  static char function[] = "gt_xio_create_handle";
  globus_result_t result;
  int status = TCL_OK;
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
  mainChannelStruct*  mainChannel = Channel->m_main_channel;
  SGDEBUG_LVL_FMT_1( 30, "mainChannel: %p", mainChannel );
  
  channelInfoStruct*  channelInfo = Channel->m_channel_info;
  SGDEBUG_LVL_FMT_1( 30, "channelInfo: %p", channelInfo );
  
  status = gt_xio_init( Channel );
  if ( status != TCL_OK )
    {
      return TCL_ERROR;
    }
  
  if ( MyPort != 0 )
    {
      result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                     channelInfo->m_xio_tcp_driver,
                                     GLOBUS_XIO_TCP_SET_CONNECT_RANGE,
                                     MyPort,
                                     MyPort );
      SGDEBUG_LVL_FMT_1( 20, "globus_xio_attr_cntl: GLOBUS_XIO_TCP_SET_CONNECT_RANGE: result: %d", result );
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
          if ( status != TCL_OK )
            {
              return TCL_ERROR;
            }
        }

      result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                     channelInfo->m_xio_tcp_driver,
                                     GLOBUS_XIO_TCP_SET_REUSEADDR,
                                     GLOBUS_TRUE );
      SGDEBUG_LVL_FMT_1( 20, "globus_xio_attr_cntl: GLOBUS_XIO_TCP_SET_REUSEADDR: result: %d",
                         result );
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
          if ( status != TCL_OK )
            {
              return TCL_ERROR;
            }
        }
    }

  result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                 channelInfo->m_xio_tcp_driver,
                                 GLOBUS_XIO_TCP_SET_NO_IPV6,
                                 GLOBUS_TRUE );
  SGDEBUG_LVL_FMT_1( 20, "globus_xio_attr_cntl: GLOBUS_XIO_TCP_SET_NO_IPV6: result: %d",
                     result );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }

  result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                 channelInfo->m_xio_tcp_driver,
                                 GLOBUS_XIO_TCP_SET_NODELAY,
                                 GLOBUS_TRUE );
  SGDEBUG_LVL_FMT_1( 20, "globus_xio_attr_cntl: GLOBUS_XIO_TCP_SET_NODELAY: result: %d",
                     result );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }


  /*
   * Setting timeout value in case we have GSI-based server with
   * non GSI-based communication channel and vice versa.
   */
 
  int secs;
  int usecs;

  /* If application specifies a timeout, use the new values */
  if ( Channel->m_user_timeout == 0 )
    {
      GlobusTimeReltimeSet(Channel->m_xio_timeout,
                           SEC_THRESHHOLD, USEC_THRESHHOLD);
    }
  else
    {
      GlobusTimeReltimeSet(Channel->m_xio_timeout,
                           Channel->m_user_timeout, 0);
    }
  
  GlobusTimeReltimeGet(Channel->m_xio_timeout, secs, usecs);
  SGDEBUG_LVL_FMT_2( 20, "timeout values: secs: %d, usecs: %d ",
                     secs, usecs );
  result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                 NULL,
                                 GLOBUS_XIO_ATTR_SET_TIMEOUT_ALL,
                                 NULL,
                                 &(Channel->m_xio_timeout),
                                 NULL );
  
  SGDEBUG_LVL_FMT_1
    ( 20, "globus_xio_attr_cntl: GLOBUS_XIO_ATTR_SET_TIMEOUT_ALL: result: %d",
      result );

  
  if ( Channel->m_gsi_auth_enabled == TRUE )
    {
      if ( Channel->m_service_name )
        {
          /*
           *=================================================================
           * This part is to handle service-based credentials
           *=================================================================
           */
          result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                         channelInfo->m_xio_gsi_driver,
                                         GLOBUS_XIO_GSI_SET_CREDENTIAL,
                                         GSS_C_NO_CREDENTIAL );
          SGDEBUG_LVL_FMT_1( 20, "globus_xio_attr_cntl: GLOBUS_XIO_GSI_SET_CREDENTIAL: result: %d",
                             result );
          if ( result != GLOBUS_SUCCESS )
            {
              status = gt_xio_check_error( result, mainChannel->m_main_interp );
              if ( status != TCL_OK )
                {
                  return TCL_ERROR;
                }
            }

          result =
            globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                  channelInfo->m_xio_gsi_driver,
                                  GLOBUS_XIO_GSI_SET_DELEGATION_MODE,
                                  Channel->m_delegation_mode );
          SGDEBUG_LVL_FMT_1
            ( 20, "GLOBUS_XIO_GSI_SET_DELEGATION_MODE: result: %d",
              result );
                    
          if ( result != GLOBUS_SUCCESS )
            {
              status = gt_xio_check_error( result, mainChannel->m_main_interp );
              if ( status != TCL_OK )
                {
                  return TCL_ERROR;
                }
            }
          
          gss_buffer_desc cert_cn;
          gss_name_t target_name;
          OM_uint32 maj_stat, min_stat;
          char cn[256];
          
          if ( Channel->m_service_name )
            {
              SGDEBUG_LVL_FMT_1( 20, "service name: %s", Channel->m_service_name );
              if ( Channel->m_server_hostname )
                {
                  SGDEBUG_LVL_FMT_1( 20, "server: %s",
                                     Channel->m_server_hostname );

                  /* Check if server name is in FQDN format */
                  if ( strchr(Channel->m_server_hostname, '.') == GLOBUS_NULL )
                    {
                      char buffer[NI_MAXHOST];
                      gt_xio_get_fqdn_host_name( Channel->m_server_hostname, buffer );
                      sprintf( cn, "%s/%s", Channel->m_service_name, buffer );
                    }
                  else
                    {
                      sprintf( cn, "%s/%s", Channel->m_service_name,
                               Channel->m_server_hostname );
                    }
                  
                  SGDEBUG_LVL_FMT_1( 20, "cn: %s", cn );
                }
              else
                {
                  /* :TODO: Throws an error */
                  return TCL_ERROR;
                }
            }
          else
            {
              if ( Channel->m_server_hostname )
                {
                  sprintf( cn, "%s", Channel->m_server_hostname );
                  SGDEBUG_LVL_FMT_1( 20, "cn: %s", cn );
                }
              else
                {
                  /* :TODO: Throws an error */
                  return TCL_ERROR;
                }
            }
          
          cert_cn.value = strdup( cn ); 
          cert_cn.length = strlen(cn) + 1;
          
          maj_stat = gss_import_name( &min_stat,
                                      &cert_cn,
                                      (gss_OID) GSS_C_NT_HOSTBASED_SERVICE,
                                      &target_name );
          
          result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                         channelInfo->m_xio_gsi_driver,
                                         GLOBUS_XIO_GSI_SET_TARGET_NAME,
                                         target_name );
          SGDEBUG_LVL_FMT_1
            ( 20,
              "GLOBUS_XIO_GSI_SET_TARGET_NAME: result: %d",
              result );
          if ( result != GLOBUS_SUCCESS )
            {      
              status = gt_xio_check_error( result, mainChannel->m_main_interp );
              if ( status != TCL_OK )
                {
                  return TCL_ERROR;
                }
            }
          
          result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                         channelInfo->m_xio_gsi_driver,
                                         GLOBUS_XIO_GSI_SET_AUTHORIZATION_MODE,
                                         GLOBUS_XIO_GSI_IDENTITY_AUTHORIZATION );
          SGDEBUG_LVL_FMT_1
            ( 20,
              "GLOBUS_XIO_GSI_SET_AUTHORIZATION_MODE: result: %d",
              result );
          if ( result != GLOBUS_SUCCESS )
            {
              status = gt_xio_check_error( result, mainChannel->m_main_interp );
              if ( status != TCL_OK )
                {
                  return TCL_ERROR;
                }
            }
          
          free( cert_cn.value );
        }
      
      else
        {
          /*
           *=================================================================
           * This code is to handle host-based credentials
           *=================================================================
           */
          result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                         channelInfo->m_xio_gsi_driver,
                                         GLOBUS_XIO_GSI_SET_CREDENTIAL,
                                         GSS_C_NO_CREDENTIAL );
          
          SGDEBUG_LVL_FMT_1
            ( 20, "GLOBUS_XIO_GSI_SET_CREDENTIAL: result: %d",
              result );
          
          if ( result != GLOBUS_SUCCESS )
            {
              status = gt_xio_check_error( result,
                                           mainChannel->m_main_interp );
              if ( status != TCL_OK )
                {
                  return TCL_ERROR;
                }
            }
          
          if ( host_authorization_mode == FALSE )
            {
              result =
                globus_xio_attr_cntl
                ( channelInfo->m_xio_handle_attr,
                  channelInfo->m_xio_gsi_driver,
                  GLOBUS_XIO_GSI_SET_AUTHORIZATION_MODE,
                  GLOBUS_XIO_GSI_HOST_AUTHORIZATION );
              
              SGDEBUG_LVL_FMT_1
                ( 20,
                  "GLOBUS_XIO_GSI_SET_AUTHORIZATION_MODE: result: %d",
                  result );
              if ( result != GLOBUS_SUCCESS )
                {
                  status = gt_xio_check_error( result, 
                                               mainChannel->m_main_interp );
                  if ( status != TCL_OK )
                    {
                      return TCL_ERROR;
                    }
                }
            }
        }
      
      /*
       * Encryption - Set the channel mode to use wrapping
       */
      result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                     channelInfo->m_xio_gsi_driver,
                                     GLOBUS_XIO_GSI_SET_WRAP_MODE,
                                     GLOBUS_TRUE );
      SGDEBUG_LVL_FMT_1( 20, "GLOBUS_XIO_GSI_SET_WRAP_MODE: result: %d",
                         result );
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
          if ( status != TCL_OK )
            {
              return TCL_ERROR;
            }
        }
      
      /*
       * Encryption - Messages are signed
       */
      result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                     channelInfo->m_xio_gsi_driver,
                                     GLOBUS_XIO_GSI_SET_PROTECTION_LEVEL,
                                     GLOBUS_XIO_GSI_PROTECTION_LEVEL_PRIVACY );
      SGDEBUG_LVL_FMT_1
        ( 10, "GLOBUS_XIO_GSI_SET_PROTECTION_LEVEL: result: %d", result );
      
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
          if ( status != TCL_OK )
            {
              return TCL_ERROR;
            }
        }
    }

#if 0
  Channel->m_xio_handle =
    (globus_xio_handle_t*)ckalloc( sizeof( globus_xio_handle_t) );
  
  result = globus_xio_handle_create( Channel->m_xio_handle,
                                     channelInfo->m_xio_stack );
#else  
  result = globus_xio_handle_create( &(Channel->m_xio_handle),
                                     channelInfo->m_xio_stack );
#endif
  
  SGDEBUG_LVL_FMT_1( 20, "globus_xio_handle_create: result: %d", result );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  
  /* Completed successfully */
  return TCL_OK;  
}

/*-----------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------*/
static int
gt_xio_connect_handle( channelStruct* Channel, int Port )
{
  static char function[] = "gt_xio_connect_handle";
  globus_result_t result;
  int status = TCL_OK;
  char contact_string[256];
  globus_size_t nbytes;
  int secs;
  int usecs;
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  mainChannelStruct*  mainChannel = Channel->m_main_channel;
  SGDEBUG_LVL_FMT_1( 30, "mainChannel: %p", mainChannel );
  
  channelInfoStruct*  channelInfo = Channel->m_channel_info;
  SGDEBUG_LVL_FMT_1( 30, "channelInfo: %p", channelInfo );
  
  /* Construct contact string from Host and Port */
  if ( Channel->m_server_hostname )
    {
      sprintf( contact_string, "%s:%d", Channel->m_server_hostname, Port );
      SGDEBUG_LVL_FMT_1( 20, "contact string: %s", contact_string );
    }
  else
    {
      Tcl_AppendResult( mainChannel->m_main_interp, 
                        "Contact string is empty",
                        NULL );
      return TCL_ERROR;
    }

  result = globus_xio_open( Channel->m_xio_handle,
                            contact_string,
                            channelInfo->m_xio_handle_attr );
  
  SGDEBUG_LVL_FMT_1( 20, "globus_xio_open: result: %d", result );
  
  if ( result != GLOBUS_SUCCESS )
    {
      if ( result == 14 || result == 15 )
        {
          Tcl_AppendResult
            ( mainChannel->m_main_interp, 
              "\nPlease check the following:\n",
              "\tClient side if using service certificates:\n",
              "\t\tMake sure you define a Tcl variable called SERVICE_NAME\n",
              "\tServer side:\n",
              "\t\tMake sure ldaskey.pem and ldascert.pem have been ",
              "installed\n",
              "\t\tin the default location\n",
              "\t\tIf certificates are not installed in the default location,\n",
              "\t\tthen please set X509_USER_KEY and X509_USER_CERT to the\n",
              "\t\tcorrect location.\n",
              "\t\tMake sure server has trusted CA certificate directory\n",
              "\t\tin the default location, otherwise set X509_CERT_DIR\n",
              "\t\tto the correct location.\n",
              "You can also check this page, http://tclglobus.ligo.caltech.edu/globus-socket.html\n",
              "for further instructions on how to diagnose the problem",
              NULL );
          
          return TCL_ERROR;
        }
      else
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp ); 
          if ( status != TCL_OK ) 
            {
              return TCL_ERROR;
            }
        }
    } 

  /* Send out byte ordering information */
  if ( gt_xio_send_byte_ordering_info( Channel ) != TCL_OK )
    {
      return TCL_ERROR;
    }

  /* Receive byte ordering information */
  if ( gt_xio_recv_byte_ordering_info( Channel ) != TCL_OK )
    {
      return TCL_ERROR;
    }

  /* turn off the timeout */
  GlobusTimeReltimeCopy(Channel->m_xio_timeout, globus_i_reltime_infinity);
  GlobusTimeReltimeGet(Channel->m_xio_timeout, secs, usecs);
  SGDEBUG_LVL_FMT_2( 20, "new timeout values: secs: %d, usecs: %d ",
                     secs, usecs );
  result = globus_xio_handle_cntl( Channel->m_xio_handle,
                                   NULL,
                                   GLOBUS_XIO_ATTR_SET_TIMEOUT_ALL,
                                   NULL,
                                   &(Channel->m_xio_timeout),
                                   NULL );
  SGDEBUG_LVL_FMT_1
    ( 20, "globus_xio_handle_cntl: GLOBUS_XIO_ATTR_SET_TIMEOUT_ALL: result: %d",
      result );
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );

  /* Completed successfully */
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int
gt_xio_create_server( channelStruct* Channel, int Port )
{
  static char function[] = "gt_xio_create_server";
  globus_result_t result;
  int status = TCL_OK;
  char* server_cs = NULL;
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  mainChannelStruct*  mainChannel = Channel->m_main_channel;
  SGDEBUG_LVL_FMT_1( 30, "mainChannel: %p", mainChannel );
  
  channelInfoStruct*  channelInfo = Channel->m_channel_info;
  SGDEBUG_LVL_FMT_1( 30, "channelInfo: %p", channelInfo );
  
  status = gt_xio_init( Channel );
  if ( status != TCL_OK )
    {
      return TCL_ERROR;
    }

  result = globus_xio_attr_init( &(mainChannel->m_xio_server_attr) );
  SGDEBUG_LVL_FMT_1( 20, "globus_xio_attr_init: result: %d", result );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }

  if ( Port != 0 )
    {
      result = globus_xio_attr_cntl( mainChannel->m_xio_server_attr,
                                     channelInfo->m_xio_tcp_driver,
                                     GLOBUS_XIO_TCP_SET_PORT,
                                     Port );
      SGDEBUG_LVL_FMT_1( 20, "GLOBUS_XIO_TCP_SET_PORT: result: %d", result );
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
          if ( status != TCL_OK )
            {
              return TCL_ERROR;
            }
        }
    }
  result = globus_xio_attr_cntl( mainChannel->m_xio_server_attr,
                                 channelInfo->m_xio_tcp_driver,
                                 GLOBUS_XIO_TCP_SET_NO_IPV6,
                                 GLOBUS_TRUE );
  SGDEBUG_LVL_FMT_1( 20, "GLOBUS_XIO_TCP_SET_NO_IPV6: result: %d", result );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }

  result = globus_xio_attr_cntl( mainChannel->m_xio_server_attr,
                                 channelInfo->m_xio_tcp_driver,
                                 GLOBUS_XIO_TCP_SET_NODELAY,
                                 GLOBUS_TRUE );
  SGDEBUG_LVL_FMT_1( 20, "GLOBUS_XIO_TCP_SET_NODELAY: result: %d", result );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }

  result = globus_xio_attr_cntl( mainChannel->m_xio_server_attr,
                                 channelInfo->m_xio_tcp_driver,
                                 GLOBUS_XIO_TCP_SET_KEEPALIVE,
                                 GLOBUS_TRUE );
  SGDEBUG_LVL_FMT_1( 20, "GLOBUS_XIO_TCP_SET_KEEPALIVE: result: %d", result );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }

  result = globus_xio_attr_cntl( mainChannel->m_xio_server_attr,
                                 channelInfo->m_xio_tcp_driver,
                                 GLOBUS_XIO_TCP_SET_REUSEADDR,
                                 GLOBUS_TRUE );
  SGDEBUG_LVL_FMT_1( 20, "GLOBUS_XIO_TCP_SET_REUSEADDR: result: %d", result );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }

  result = globus_xio_attr_cntl( mainChannel->m_xio_server_attr,
                                 channelInfo->m_xio_tcp_driver,
                                 GLOBUS_XIO_GSI_FORCE_SERVER_MODE,
                                 GLOBUS_TRUE );
  SGDEBUG_LVL_FMT_1( 20, "GLOBUS_XIO_GSI_FORCE_SERVER_MODE: result: %d",
                     result );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }

  if ( Channel->m_gsi_auth_enabled == TRUE )
    {
      result = globus_xio_attr_cntl( mainChannel->m_xio_server_attr,
                                     channelInfo->m_xio_gsi_driver,
                                     GLOBUS_XIO_GSI_FORCE_SERVER_MODE,
                                     GLOBUS_TRUE );
      SGDEBUG_LVL_FMT_1( 20, "GLOBUS_XIO_GSI_FORCE_SERVER_MODE: result: %d",
                         result );
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
          if ( status != TCL_OK )
            {
              return TCL_ERROR;
            }
        }

      /*
       * Encryption - Set the channel mode to use wrapping
       */
      result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                     channelInfo->m_xio_gsi_driver,
                                     GLOBUS_XIO_GSI_SET_WRAP_MODE,
                                     GLOBUS_TRUE );
      SGDEBUG_LVL_FMT_1( 20, "GLOBUS_XIO_GSI_SET_WRAP_MODE: result: %d",
                         result );
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
          if ( status != TCL_OK )
            {
              return TCL_ERROR;
            }
        }
      
      /*
       * Encryption - Messages are signed
       */
      result = globus_xio_attr_cntl( channelInfo->m_xio_handle_attr,
                                     channelInfo->m_xio_gsi_driver,
                                     GLOBUS_XIO_GSI_SET_PROTECTION_LEVEL,
                                     GLOBUS_XIO_GSI_PROTECTION_LEVEL_PRIVACY );
      SGDEBUG_LVL_FMT_1( 20, "GLOBUS_XIO_GSI_SET_PROTECTION_LEVEL: result: %d",
                         result );
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
          if ( status != TCL_OK )
            {
              return TCL_ERROR;
            }
        }
    }
  
  /* Create Globus XIO server */
  result = globus_xio_server_create( &(mainChannel->m_xio_server),
                                     mainChannel->m_xio_server_attr,
                                     channelInfo->m_xio_stack );
  SGDEBUG_LVL_FMT_1( 20, "globus_xio_server_create: result: %d", result );
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }

  result = globus_xio_server_get_contact_string( mainChannel->m_xio_server,
                                                 &server_cs );
  SGDEBUG_LVL_FMT_1( 20, "globus_xio_server_get_contact_string: result: %d",
                     result );
  SGDEBUG_LVL_FMT_1( 20, "globus_xio_server_get_contact_string: server_cs: %s",
                     server_cs );
  globus_free(server_cs);
  
  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );

  /* Completed successfully */
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 * Callback routines
 *-----------------------------------------------------------------------*/
static int
gt_xio_get_peer_info( channelStruct* Channel,
                      char *PeerHostname,
                      char* PeerIPAddr,
                      int* PeerPort )
{
  static char function[] = "gt_xio_get_peer_info";
  globus_result_t result;
  int status = TCL_OK; 
  char* contact;
  char* hostname;
  char* ipaddr;
  char* port;
  char* lasts;
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
  SGDEBUG_LVL_FMT_1( 30, "Channel: %p", Channel );
  
  mainChannelStruct*  mainChannel = Channel->m_main_channel;
  SGDEBUG_LVL_FMT_1( 30, "mainChannel: %p", mainChannel );
  
  channelInfoStruct*  channelInfo = Channel->m_channel_info;
  SGDEBUG_LVL_FMT_1( 30, "channelInfo: %p", channelInfo );
  
  result = globus_xio_handle_cntl( Channel->m_xio_handle,
                                   channelInfo->m_xio_tcp_driver,
                                   GLOBUS_XIO_TCP_GET_REMOTE_CONTACT,
                                   &contact );

  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }

  SGDEBUG_LVL_FMT_1( 20, "GLOBUS_XIO_TCP_GET_REMOTE_CONTACT: contact: %s",
                     contact );
  
  /* Get peer hostname */
  /* hostname = strtok( contact, ":" ); */
  hostname = strtok_r( contact, ":", &lasts );
  sprintf( PeerHostname, "%s", hostname );
    
  /* Get peer port number */
  /* port = strtok( NULL, ":" ); */
  port = strtok_r( NULL, ":", &lasts );
  *PeerPort = atoi( port );

  result =
    globus_xio_handle_cntl( Channel->m_xio_handle,
                            channelInfo->m_xio_tcp_driver,
                            GLOBUS_XIO_TCP_GET_REMOTE_NUMERIC_CONTACT,
                            &contact );

  if ( result != GLOBUS_SUCCESS )
    {
      status = gt_xio_check_error( result, mainChannel->m_main_interp );
      if ( status != TCL_OK )
        {
          return TCL_ERROR;
        }
    }

  SGDEBUG_LVL_FMT_1
    ( 20, "GLOBUS_XIO_TCP_GET_REMOTE_NUMERIC_CONTACT: contact: %s", contact );

  /* Get peer IP address */
  /* ipaddr = strtok( contact, ":" ); */
  ipaddr = strtok_r( contact, ":", &lasts );
  sprintf( PeerIPAddr, "%s", ipaddr );
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );

  /* Completed successfully */
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static void
accept_callback_proc( ClientData Data, 
                      char* PeerHostname, 
                      int PeerPort )
{
  static char function[] = "accept_callback_proc";
  int status = TCL_OK;
  char command[1024];
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  channelStruct* channelPtr = (channelStruct*)Data;
  SGDEBUG_LVL_FMT_1( 30, "channelPtr: %p", channelPtr );

  mainChannelStruct*  mainChannel = channelPtr->m_main_channel;
  SGDEBUG_LVL_FMT_1( 30, "mainChannel: %p", mainChannel );
  
  if ( mainChannel->m_main_interp != (Tcl_Interp*)NULL )
    {
      SGDEBUG_LVL_FMT( 20, "Prior to calling Tcl_Preserve" );
      /* Protect ourselves */
      SGDEBUG_LVL_FMT_1( 20, "m_script: %s", channelPtr->m_script );
      SGDEBUG_LVL_FMT_1( 20, "m_script: %p", channelPtr->m_script );
      
      SGDEBUG_LVL_FMT_1( 20, "m_main_interp: %p",
                         mainChannel->m_main_interp );
      
      Tcl_Preserve( (ClientData)( channelPtr->m_script ) );
      Tcl_Preserve( (ClientData)( mainChannel->m_main_interp ) );
      
      /* Invoke any script */
      sprintf( command, "%s %s %s %d",
               channelPtr->m_script,
               Tcl_GetChannelName( channelPtr->m_channel ),
               PeerHostname,
               PeerPort );
            
      SGDEBUG_LVL_FMT_1( 20, "command %s", command );

      status = Tcl_VarEval( mainChannel->m_main_interp,
                            command,
                            (char*)NULL );
      
      /* Something is wrong */
      if ( status != TCL_OK )
        {
          SGDEBUG_HERE( );
          SGDEBUG_LVL_FMT_1( 20, "main interp %p",
                             mainChannel->m_main_interp );

          Tcl_BackgroundError( mainChannel->m_main_interp );

          Tcl_UnregisterChannel( mainChannel->m_main_interp,
                                 channelPtr->m_channel );
        }
      
      /* Release resources */
      Tcl_Release( (ClientData)( channelPtr->m_script ) );
      Tcl_Release( (ClientData)( mainChannel->m_main_interp ) );
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static void
xio_accept( ClientData Data, int Mask )
{
  static char function[] = "xio_accept";

  globus_result_t result;
  int status = TCL_OK;
  char peer_hostname[256];
  char peer_ipaddr[256];
  int peer_port;
  char* server_cs = NULL;

  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
  channelStruct* channelPtr = (channelStruct*)Data;
  SGDEBUG_LVL_FMT_1( 30, "channelPtr: %p", channelPtr );

  mainChannelStruct* mainChannel = channelPtr->m_main_channel;
  SGDEBUG_LVL_FMT_1( 30, "mainChannel: %p", mainChannel );
  
  channelInfoStruct*  channelInfo = channelPtr->m_channel_info;
  SGDEBUG_LVL_FMT_1( 30, "channelInfo: %p", channelInfo );
  
  SGDEBUG_LVL_FMT_1( 20, "m_main_interp: %p",
                     mainChannel->m_main_interp );
  
  if ( mainChannel->m_main_interp != (Tcl_Interp*)NULL )
    {
      SGDEBUG_LVL_FMT_1( 20, "m_script: %s", channelPtr->m_script );

      /*================================================================*/
      /* Allocate channel space for the new client 			*/
      /*================================================================*/
      channelStruct* newClientChannel =
        (channelStruct*)globus_libc_malloc( sizeof(channelStruct) );
      SGDEBUG_LVL_FMT_1( 20, "new client channel: %p", newClientChannel );

      newClientChannel->m_main_channel = channelPtr->m_main_channel;
      newClientChannel->m_channel_info = channelPtr->m_channel_info;

      newClientChannel->m_script = strdup( channelPtr->m_script );
      newClientChannel->m_delete_on_close = TRUE;
      newClientChannel->m_gsi_auth_enabled = channelPtr->m_gsi_auth_enabled;
      
      /*================================================================*/
      /* Accept incoming calls 						*/
      /*================================================================*/
      result = globus_xio_server_accept( &(newClientChannel->m_xio_handle),
                                         mainChannel->m_xio_server );
      
      SGDEBUG_LVL_FMT_1( 20, "globus_xio_server_accept: result: %d",
                         result );
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
          if ( status != TCL_OK )
            {
              return;
            }
        }
      
      SGDEBUG_LVL_FMT_1( 20, "channelPtr->m_xio_handle: result: %p",
                         newClientChannel->m_xio_handle );
      
      /*================================================================*/
      /* Check for X509_USER_KEY and X509_USER_CERT values		*/
      /*================================================================*/
      if ( channelPtr->m_gsi_auth_enabled == TRUE )
        {
          /* Checking for X.509 cert directory */
          status = gt_xio_check_x509_cert_dir( mainChannel->m_main_interp );
          if ( status != TCL_OK )
            {
              return;
            }

          /* Checking for X.509 user cert and user key variables */
          result = gt_xio_check_x509( mainChannel->m_main_interp );
          if ( result != TCL_OK )
            {
              return;
            }
        }
      
      /*================================================================*/
      /* Handle is initialized and it's time to open for read/write	*/
      /*================================================================*/
      result = 
        globus_xio_server_get_contact_string( mainChannel->m_xio_server,
                                              &server_cs );
      SGDEBUG_LVL_FMT_1
        ( 20, "globus_xio_server_get_contact_string: result: %d",
          result );
      SGDEBUG_LVL_FMT_1
        ( 20, "globus_xio_server_get_contact_string: server_cs: %s",
          server_cs );

      result =
        globus_xio_open( newClientChannel->m_xio_handle,
                         server_cs,
                         newClientChannel->m_channel_info->m_xio_handle_attr );

      SGDEBUG_LVL_FMT_1( 20, "globus_xio_open: result: %d", result );
      if ( result != GLOBUS_SUCCESS )
        {
          status = gt_xio_check_error( result, mainChannel->m_main_interp );
          if ( status != TCL_OK )
            {
              return;
            }
        }
      globus_free( server_cs );
            
      /*================================================================*/
      /* Create a new client channel 					*/
      /*================================================================*/
      newClientChannel->m_channel =
        Tcl_CreateChannel( get_channel_type( ),
                           gen_socket_name( ),
                           (ClientData)newClientChannel,
                           (TCL_READABLE | TCL_WRITABLE) );

      Tcl_RegisterChannel( mainChannel->m_main_interp,
                           newClientChannel->m_channel );
        
      /*================================================================*/
      /* Get peer information 						*/
      /*================================================================*/
      status = gt_xio_get_peer_info( newClientChannel, peer_hostname,
                                     peer_ipaddr, &peer_port );
      if ( status != TCL_OK )
        {
          return;
        }
      
      SGDEBUG_LVL_FMT_1( 20, "peer_hostname: %s", peer_hostname );
      SGDEBUG_LVL_FMT_1( 20, "peer_ipaddr: %s", peer_ipaddr );
      SGDEBUG_LVL_FMT_1( 20, "peer_port: %d", peer_port );
      
      /*================================================================*/
      /* Get byte ordering information 					*/
      /*================================================================*/
      if ( gt_xio_recv_byte_ordering_info( newClientChannel ) != TCL_OK )
        {
          return;
        }

      /*================================================================*/
      /* Send out byte ordering information 				*/
      /*================================================================*/
      if ( gt_xio_send_byte_ordering_info( newClientChannel ) != TCL_OK )
        {
          return;
        }
      
      /*================================================================*/
      /* Do the callback thing 						*/
      /*================================================================*/
      accept_callback_proc( (ClientData)newClientChannel,
                            peer_hostname,
                            peer_port );

    }

  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static void
delete_proc( ClientData Data, Tcl_Interp* Interp )
{
  static char function[] = "delete_proc";
  Tcl_HashEntry* hPtr;
  Tcl_HashSearch hSearch;
  Tcl_HashTable* hTblPtr;
  channelStruct* channelPtr;
                                  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  hTblPtr = (Tcl_HashTable*)Data;

  for ( hPtr = Tcl_FirstHashEntry( hTblPtr, &hSearch );
        hPtr != (Tcl_HashEntry*)NULL;
	hPtr = Tcl_NextHashEntry( &hSearch ) )
    {
      channelPtr = (channelStruct*)Tcl_GetHashValue( hPtr );
      channelPtr->m_main_channel->m_main_interp = (Tcl_Interp*)NULL;
    }
  Tcl_DeleteHashTable( hTblPtr );
  ckfree( (char*)hTblPtr );
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static void
server_close_proc( ClientData Data )
{
  static char function[] = "server_close_proc";
  Tcl_HashTable* hTblPtr;
  Tcl_HashEntry* hPtr;

  channelStruct* channelPtr = (channelStruct*)Data;
  mainChannelStruct* mainChannel = channelPtr->m_main_channel;
  channelInfoStruct* channelInfo = channelPtr->m_channel_info;
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  if ( mainChannel->m_server_mode == FALSE )
    {
      return;
    }
  if ( mainChannel->m_main_interp !=  (Tcl_Interp*)NULL )
    {
      hTblPtr =
        (Tcl_HashTable*)Tcl_GetAssocData( mainChannel->m_main_interp,
                                          HASH_KEY,
                                          NULL );
      if ( hTblPtr == (Tcl_HashTable*)NULL )
        {
          return;
        }
      hPtr = Tcl_FindHashEntry(hTblPtr, (char*)channelPtr );
      if ( hPtr == (Tcl_HashEntry*)NULL )
        {
          return;
        }
      Tcl_DeleteHashEntry( hPtr );
      mainChannel->m_main_interp = (Tcl_Interp*)NULL;
    }
  
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
}
