/* -*- indent-tabs-mode: nil; -*- */
#include "config.h"

#include <stdio.h>
#include <pthread.h>
#include <stdarg.h>
#include <strings.h>

#include <tcl.h>

#include "common/sgdebug.h"

#include "tcl-callback.h"
#include "globus_libc.h"
#include "globus_xio_types.h"

#if SIZEOF_INT == 4
typedef int INT_4S;
typedef unsigned int INT_4U;
#elif SIZEOF_LONG == 4
typedef long INT_4S;
typedef unsigned long INT_4U;
#else
#   error "No 4 byte integer found"
#endif

#if SIZEOF_LONG == 8
typedef long INT_8S;
typedef unsigned long INT_8U;
#elif SIZEOF_LONG_LONG == 8
typedef long long INT_8S;
typedef unsigned long long INT_8U;
#else
#   error "No 8 byte integer found"
#endif

#if SIZEOF_VOIDP == 4
typedef INT_4U MEM_PTR;
#elif SIZEOF_VOIDP == 8
typedef INT_8U MEM_PTR;
#else
#   error "Unexpected sizeof(void*)"
#endif
#define SIZEOF_MEM_PTR SIZEOF_VOIDP

static struct {
  pthread_mutex_t	lock;
  unsigned int		length;
  unsigned int		capacity;
  callbackStruct**	queue;
} tg_callback_queue = {
  PTHREAD_MUTEX_INITIALIZER,	/* lock */
  0,				/* length */
  0,				/* capacity */
  (callbackStruct**)NULL	/* queue */
};

static int IsTk;

static int PAUSE_SEC = 0;
static int PAUSE_USEC = 100;

static void callback_queue_append( callbackStruct* Callback );
static void callback_queue_process( );

void
TclGlobusSetMaxBlockTime( int Seconds, int USeconds )
{
  PAUSE_SEC = Seconds;
  PAUSE_USEC = USeconds;
}
void
TclGlobusCreateUniqueTCLVariableName( char* Buffer )
{
  struct timeval time;
  int retval;
  static char function[] = "TclGlobusCreateUniqueTCLVariableName";

  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
  retval = gettimeofday(&time, NULL);
  SGDEBUG_LVL_FMT_1( 10, "gettimeofday returns: %d", retval );
  if ( retval != 0 )
    {
      fprintf(stderr, "gettimeofday fails");
      exit(1);
    }
  sprintf( Buffer, "::tcl_globus_tmp%d%06d", time.tv_sec, time.tv_usec );
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
}

static void globus_event_check( ClientData, int Flags );
static  int globus_event_handler( Tcl_Event* Event, int Flags );
static void globus_event_setup( ClientData, int Flags );

static  int
globus_event_handler( Tcl_Event* Event, int Flags )
{
  static char function[] = "globus_event_handler";
  SGDEBUG_LVL_FMT_1( 30, "Entering %s", function );
  
#if ! HAVE_GLOBUS_THREADED
  //-------------------------------------------------------------------
  // Process any globus events
  //-------------------------------------------------------------------
  globus_poll( );
#endif /* ! HAVE_GLOBUS_THREADED */

  //-------------------------------------------------------------------
  // Process any TclGlobus callbacks
  //-------------------------------------------------------------------
  callback_queue_process( );

  //-------------------------------------------------------------------
  // Setup for next round of changes
  //-------------------------------------------------------------------
  //globus_event_setup( NULL, 0 );

  SGDEBUG_LVL_FMT_1( 30, "Leaving %s", function );
  return 1;
}

static void
globus_event_check( ClientData CD, int Flags )
{
  static char function[] = "globus_event_check";
  Tcl_Event*	event;

  SGDEBUG_LVL_FMT_1( 30, "Entering %s", function );
  
#if ! HAVE_GLOBUS_THREADED
  //-------------------------------------------------------------------
  // Process any globus events
  //-------------------------------------------------------------------
  globus_poll( );
#endif /* ! HAVE_GLOBUS_THREADED */
  
  if ( ( ! IsTk ) || ( tg_callback_queue.length ) )
  {
    event = (Tcl_Event*)(ckalloc( sizeof( Tcl_Event ) ) );
      
    event->proc = globus_event_handler;
      
    Tcl_QueueEvent( event, TCL_QUEUE_TAIL );
  }
  SGDEBUG_LVL_FMT_1( 30, "Leaving %s", function );
}

static void
globus_event_setup( ClientData CD, int Flags )
{
  Tcl_Time	timeout;
  static char function[] = "globus_event_setup";
  
  SGDEBUG_LVL_FMT_1( 30, "Entering %s", function );
  
#if ! HAVE_GLOBUS_THREADED
  //-------------------------------------------------------------------
  // Process any globus events
  //-------------------------------------------------------------------
  globus_poll( );
#endif /* ! HAVE_GLOBUS_THREADED */
  
  if ( tg_callback_queue.length )
  {
    timeout.sec = 0;
    timeout.usec = 0;
  }
  else
  {
    timeout.sec = PAUSE_SEC;
    timeout.usec = PAUSE_USEC;
  }
  Tcl_SetMaxBlockTime( &timeout );
  
  SGDEBUG_LVL_FMT_1( 30, "Leaving %s", function );
}

char*
makeTclPointer( const void* p,
		const char* classname,
		char* output )
{
  //SGDEBUG_FMT_1( "mem: %p", p );
  MEM_PTR	mem = (MEM_PTR)( p );
#if WORDS_BIGENDIAN != 1
  unsigned char tmp;
  unsigned char* d = (unsigned char*)&mem;
#if SIZEOF_MEM_PTR == 4
  tmp = *d;
  *d = *( d + 3 );
  *( d + 3 ) = tmp;
  tmp = *( d + 1 );
  *( d + 1 ) = *( d + 2 );
  *( d + 2 ) = tmp;
#elif SIZEOF_MEM_PTR == 8
  tmp = *d;
  *d = *( d + 7 );
  *( d + 7 ) = tmp;
  tmp = *( d + 1 );
  *( d + 1 ) = *( d + 6 );
  *( d + 6 ) = tmp;
  tmp = *( d + 2 );
  *( d + 2 ) = *( d + 5 );
  *( d + 5 ) = tmp;
  tmp = *( d + 3 );
  *( d + 3 ) = *( d + 4 );
  *( d + 4 ) = tmp;
#endif /* sizeof(MEM_PTR) */
#endif /* WORDS_BIGENDIAN != 1 */

  sprintf( output, "_%08x_p_%s", mem, classname );
  SGDEBUG_FMT_1( "output: %s", output );
  return output;
}
/**
 * This function will create an event handler in the case of linking to
 *   non-threaded version of globus packages.
 */
static void
create_globus_event_handler( Tcl_Interp* Interp )
{
  static int initialized = 0;
  struct Tcl_CmdInfo cmdinfo;
  static char function[] = "create_globus_event_handler";

  SGDEBUG_LVL_FMT_1( 30, "Entering %s", function );
  if ( initialized == 0 )
  {
    Tcl_CreateEventSource( globus_event_setup, globus_event_check, NULL );
    IsTk = Tcl_GetCommandInfo( Interp, "wm", &cmdinfo );
    SGDEBUG_LVL_FMT_1( 10, "create_globus_event_handler: IsTk: %d", IsTk );
    initialized = 1;
  }
  SGDEBUG_LVL_FMT_1( 30, "Leaving %s", function );
}

static void
callback_preprocess( callbackStruct* Callback )
{
  static char function[] = "callback_preprocess";
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  if ( Callback->s_pre_process )
  {
    (*(Callback->s_pre_process))(Callback);
  }
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
}

static void
callback_postprocess( callbackStruct* Callback )
{
  static char function[] = "callback_postprocess";
  int cleanup = 1;

  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  SGDEBUG_LVL_FMT_1( 10, "Callback: %p", Callback );
  
  if ( Callback->s_post_process )
  {
    cleanup = (*(Callback->s_post_process))(Callback);
  }

  SGDEBUG_LVL_FMT_1( 10, "cleanup: %d", cleanup );
  
  if ( cleanup )
    {
      SGDEBUG_LVL_FMT( 10, "Cleaning up callback resources" );
    
      Tcl_DecrRefCount( Callback->func ); 
      Tcl_DecrRefCount( Callback->arg ); 
      globus_libc_free( Callback );
    }
  else
    {
      SGDEBUG_LVL_FMT( 10, "No Cleanup of resources" );
    }
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
}

void
TclGlobusCallbackEval( callbackStruct* Callback )
{
  int	result;		/* Result of Tcl evaluation			*/
  char	cmd[1024];	/* Buffer to hold user's request 		*/
  static char function[] = "TclGlobusCallbackEval";

  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  SGDEBUG_LVL_FMT_1( 10, "Callback %p", Callback );
  
  /*---------------------------------------------------------------------
   * Do any preprocessing
   *---------------------------------------------------------------------*/
  callback_preprocess( Callback );

  /* Evaluate Caller's request */
  sprintf( cmd,
	   "%s %s %s\n",
	   Tcl_GetString( Callback->func ),
	   Tcl_GetString( Callback->arg ),
	   Callback->s_vargs );

  SGDEBUG_LVL_FMT_1( 10, "Tcl full command: '%s'", cmd );

  result = Tcl_Eval( Callback->s_main_tcl_interp, cmd );

  SGDEBUG_LVL_FMT_1( 10, "Result of Tcl_Eval: %d", result );

  if ( result != TCL_OK )
  {
    SGDEBUG_LVL_FMT( 10, "Setting up to call Tcl_BackgroundError" );

    /*-------------------------------------------------------------------
     * Something went wrong, pass the error to bgerror
     *-------------------------------------------------------------------*/
    SGDEBUG_LVL_FMT_1( 10, "Unable to execute the command: %s", cmd );
    SGDEBUG_LVL_FMT_1( 10, "Result String: %s", Tcl_GetStringResult( Callback->s_main_tcl_interp ) );
    Tcl_BackgroundError( Callback->s_main_tcl_interp );
  }
  /*---------------------------------------------------------------------
   * Do any post processing
   *---------------------------------------------------------------------*/
  callback_postprocess( Callback );

  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
}

/**
 * This function allows for a single instance of a "Callback" TCL Interpreter
 *
 * This function causes callers from multiple threads to synchronize the
 * use of the single Tcl Interpreter that is reserved for making 
 */
static void
submit_callback( callbackStruct* Callback )
{
  pthread_t	this_thread;
  static char function[] = "submit_callback";

  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
  this_thread = pthread_self( );
  SGDEBUG_LVL_FMT_3( 10, "s_main_thread_id: %p this_thread: %p =?=: %d",
                     Callback->s_main_thread_id, this_thread,
                     pthread_equal( Callback->s_main_thread_id,
                                    this_thread ) );
  
  if ( pthread_equal( Callback->s_main_thread_id, this_thread ) ) 
  {
    /*-------------------------------------------------------------------
     * Execution should be done locally
     *-------------------------------------------------------------------*/
    SGDEBUG_LVL_FMT( 10, "The globus callback is in the same thread as TCL" );
    TclGlobusCallbackEval( Callback );
  }
  else
  {
    /*-------------------------------------------------------------------
     * Execution need to happen remotely
     *-------------------------------------------------------------------*/
    SGDEBUG_LVL_FMT_1( 10, "The globus callback is in another thread from TCL: Command", Callback->s_vargs );
    callback_queue_append( Callback );
  }
  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
}

/* Generic FUNCTION */
void
CallCallback( callbackStruct* Callback, char* Format, ... )
{
  va_list	ap;
  static char function[] = "CallCallback";
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
  va_start( ap, Format );

  SGDEBUG_LVL_FMT_1( 10, "Format: '%s'", Format );

  vsprintf( Callback->s_vargs, Format, ap );

  SGDEBUG_LVL_FMT_1( 10, "vargs: '%s'", Callback->s_vargs );
  
  submit_callback( Callback );

  va_end( ap );

  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  return; 
}

/* GENERIC FUNCTION */
callbackStruct*
CreateCallback( Tcl_Interp* Interp, Tcl_Obj* Input )
{
  static initialized = 0;
  static char function[] = "CreateCallback";
  
  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );
  
  if ( initialized == 0 )
  {
    Tcl_callback_call_SafeInit( Interp );
    initialized = 1;
  }
  create_globus_event_handler( Interp );

  callbackStruct* cbStruct =
    (callbackStruct*)globus_libc_malloc(sizeof(callbackStruct)); 

  SGDEBUG_LVL_FMT_1( 10, "Allocating callback space: '%p'", cbStruct );
    
  Tcl_ListObjIndex( Interp, Input, 0, &cbStruct->func ); 
  Tcl_IncrRefCount( cbStruct->func );
  SGDEBUG_LVL_FMT_1( 10, "cbStruct->func: '%p'", cbStruct->func );
  Tcl_ListObjIndex( Interp, Input, 1, &cbStruct->arg );
  Tcl_IncrRefCount( cbStruct->arg );
  SGDEBUG_LVL_FMT_1( 10, "cbStruct->arg: '%p'", cbStruct->arg );
  cbStruct->s_main_thread_id = pthread_self( );
  cbStruct->s_main_tcl_interp = Interp;

  cbStruct->s_api_data = NULL;
  cbStruct->s_pre_process = NULL;
  cbStruct->s_post_process = NULL;

  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
  return cbStruct;
}

static void
callback_queue_append( callbackStruct* Callback )
{
  callbackStruct**	pos;
  unsigned int	x;
  static char function[] = "callback_queue_append";

  SGDEBUG_LVL_FMT_1( 10, "Entering %s", function );

  SGDEBUG_LVL_FMT_1( 10, "Callback %p", Callback );
  
  /* Mutex lock - acquire */
  pthread_mutex_lock( &(tg_callback_queue.lock) );
  
  if ( tg_callback_queue.length ==
       tg_callback_queue.capacity )
  {
    /* Allocate new memory */
    unsigned int old_capacity = tg_callback_queue.capacity;
    unsigned int new_capacity;

    new_capacity = tg_callback_queue.capacity += 5;
    tg_callback_queue.queue =
      (callbackStruct**)realloc( tg_callback_queue.queue,
                                 new_capacity * sizeof( callbackStruct* ) );
    /* Fill in the new bytes will null pointers */
    memset( &(tg_callback_queue.queue[ old_capacity ]),
            0,
            5 * sizeof( callbackStruct* ) );
  }
  pos = tg_callback_queue.queue;
  for ( x = 0; x < tg_callback_queue.capacity; ++x, ++pos )
  {
    if ( *pos == (callbackStruct*)NULL )
    {
      *pos = Callback;
      ++tg_callback_queue.length;
      break;
    }
  }

  /* Mutex lock - release */
  pthread_mutex_unlock( &(tg_callback_queue.lock) );

  SGDEBUG_LVL_FMT_1( 10, "Leaving %s", function );
}

static void
callback_queue_process( )
{
  callbackStruct**	pos;
  callbackStruct*	cb;
  unsigned int		x;
  static char function[] = "callback_queue_process";

  SGDEBUG_LVL_FMT_1( 30, "Entering %s", function );
  
  pos = tg_callback_queue.queue;
  for ( x = 0; x < tg_callback_queue.capacity; ++x, ++pos )
  {
    if ( *pos != (callbackStruct*)NULL )
    {
      /* Mutex lock - acquire */
      pthread_mutex_lock( &(tg_callback_queue.lock) );
  
      cb = *pos;
      *pos = (callbackStruct*)NULL;
      --tg_callback_queue.length;
      /* Mutex lock - release */
      pthread_mutex_unlock( &(tg_callback_queue.lock) );

      /* Execute in the tcl thread */
      TclGlobusCallbackEval( cb );
    }
  }

  SGDEBUG_LVL_FMT_1( 30, "Leaving %s", function );
}
