#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include "common/sgdebug.h"

char *
timestamp( )
{
  struct timeval t;
  static char buffer[64];
  
  gettimeofday(&t, NULL);
  sprintf( buffer, "%d.%06d", t.tv_sec, t.tv_usec );

  return ( buffer );
}

void
sbrkdebug( const char* File, const int Line )
{  
  static int prev_sbrk_value = 0;
  static int curr_sbrk_value;
  int sbrk_diff;

  static pthread_mutex_t	lock = PTHREAD_MUTEX_INITIALIZER;

  pthread_mutex_lock( &lock );	/* Mutex lock - acquire */
    
  /* Get current sbrk value */
  curr_sbrk_value = (int)((int*)sbrk( 0 ));

  if ( prev_sbrk_value == 0 )
    {
      prev_sbrk_value = curr_sbrk_value;
    }
  else
    {
      /*
       * Calculate the current sbrk difference
       */
      sbrk_diff = curr_sbrk_value - prev_sbrk_value;

      if ( sbrk_diff >= SBRK_MAX_VALUE )
	{
	  /*
	   * Print out the relevant sbrk information
	   */
	  fprintf( stderr, "DEBUG: C: previous sbrk value: %d [%s - %d]\n",
		   prev_sbrk_value, File, Line );
	  fprintf( stderr, "DEBUG: C: current sbrk value: %d [%s - %d]\n",
		   curr_sbrk_value, File, Line );
	  
	  /*
	   * Kill this process and dump a core for further analysis
	   */
	  kill( getpid( ), SIGSEGV );
	}
      else
	{
	  if ( sbrk_diff >= SBRK_THRESHOLD_VALUE )
	    {
	      /*
	       * Print out the relevant sbrk information
	       */
	      fprintf( stderr, "DEBUG: C: previous sbrk value: %d [%s - %d]\n",
		       prev_sbrk_value, File, Line );
	      fprintf( stderr, "DEBUG: C: current sbrk value: %d [%s - %d]\n",
		       curr_sbrk_value, File, Line );
	      
	      /*
	       * Flag the application for the memory increases
	       */
	      fprintf( stderr, "DEBUG: C: sbrk value diff: %d\n", sbrk_diff );
	      fprintf( stderr, "DEBUG: C: WARNING: Memory heap size has been increased by more than %d bytes\n",
		       SBRK_THRESHOLD_VALUE );

	      prev_sbrk_value = curr_sbrk_value;
	    }
	}
    }
  
  pthread_mutex_unlock( &lock );	/* Mutex lock - release */
}

void
sgdebug( int Level, const char* Format, ... )
{
  /*---------------------------------------------------------------------
   * Support for variable number of arguments
   *---------------------------------------------------------------------
   */
  va_list	ap;

  /*---------------------------------------------------------------------
   * local "constants". They will be set in the initialization section
   *---------------------------------------------------------------------
   */
  static int			debug_level = -1;
  static int			initialized = 0;
  static pthread_mutex_t	lock = PTHREAD_MUTEX_INITIALIZER;

  va_start(ap, Format);	/* Get the start of the variable arguments */
  if ( ! initialized )
  {
    /*-------------------------------------------------------------------
     * Stop everyone here
     *-------------------------------------------------------------------
     */
    pthread_mutex_lock( &lock );	/* Mutex lock - acquire */
    if ( ! initialized )
    {
      /*-----------------------------------------------------------------
       * Initialization has not yet happened
       *-----------------------------------------------------------------
       */
      const char* mode = getenv( "C_VERBOSE_MODE" );
      if ( mode )
      {
	if ( strcmp( mode, "true" ) == 0 )
	{
	  debug_level = 1;
	}
	else
	{
	  register int	x = 0;
	  register last = strlen( mode );
	  
	  while ( x < last )
	  {
	    if ( ! isdigit( mode[ x ] ) )
	    {
	      /*---------------------------------------------------------
	       * Found non-digit character, assume debugging is
	       * disabled.
	       *---------------------------------------------------------
	       */  
	      break;
	    }
	    x++;
	  }
	  if ( x == last )
	  {
	    debug_level = atoi( mode );
	  }
	}
      }
      /*-----------------------------------------------------------------
       * Inform others that initialization has happened
       *-----------------------------------------------------------------
       */
      initialized = 1;
    }
    pthread_mutex_unlock( &lock );	/* Mutex lock - release */
  }
  
  if ( debug_level >= Level )
    {
      vfprintf( stderr, Format, ap );

#if SBRK_DEBUG      
      sbrkdebug( SBRKDEBUG_SUF_DATA );
#endif      
    }
  
  va_end(ap);	/* Complete usage of variable number of arguments */
}
