#------------------------------------------------------------------------
# Routine to read the scripting language neutral resource file and
#  populate the environment variables.
#------------------------------------------------------------------------
proc load_rc {filename} {
    if { [ catch { set fid [ open $filename r ] } err ] } {
	puts_fail "Unable to load: $filename: $err"
	exit 1
    }
    set section "defaults"
    foreach line [split [read $fid] \n] {
	set line [string trim $line]
	if { [string length $line] == 0 } {
	} elseif [regexp -- {^\[(.*)\]$} $line -> section] {
	    set section [string trim $section]
	} else {
	    foreach {var value} [split $line =] {
		set var [string trim $var]
		set value [string trim $value]
		set ::TestEnv($section,$var) $value
	    }
	}
    }
    close $fid
}
proc validate_section {section vars} {
    set ok 1
    foreach var $vars {
	if {! [ info exists ::TestEnv($section,$var) ] } {
	    puts_fail "Configuration information missing in section: $section for var: $var"
	    set ok 0
	}
    }
    if { ! $ok } {
	exit 1
    }
}
#------------------------------------------------------------------------
# From: Practical Programming in Tcl and Tk 4th Edition pp. 139
# Assign a set of variables from a alist of values.
# If there are more values than variables, they are returned.
# If there are fewer values than variables,
# the variables get the empty string
#------------------------------------------------------------------------
proc lassign {valueList args} {
    if { [llength $args] == 0} {
	error "wrong \# args: lassign list varname ?varname..?"
    }
    if { [llength $valueList] == 0} {
	# Ensure one trip through the foreach loop
	set valueList [list {}]
    }
    uplevel 1 [list foreach $args $valueList {break}]
    return [lrange $valueList [llength $args] end]
}
#------------------------------------------------------------------------
# Debug procedure
#------------------------------------------------------------------------
proc Debug { args } {
    global debug
    if { ![info exists debug(enabled)] } {
	# Default is to do nothing
	return
    }
    puts $debug(file) [ join $args " " ]
}

proc DebugOn { {file {}} } {
    global debug
    set debug(enabled) 1
    if { [string length $file] == 0 } {
	set debug(file) stderr
    } else {
	if [catch {open $file w} fileID] {
	    puts stderr "Cannot open $file: $fileID"
	    set debug(file) stderr
	} else {
	    puts stderr "Debug info to $file"
	    set debug(file) $fileID
	}
    }
}

proc DebugOff {} {
    global debug
    if {[info exists debug(enabled)]} {
	unset debug(enabled)
	flush $debug(file)
	if {$debug(file) != "stderr" &&
	    $debug(file) != "stdout"} {
	    close $debug(file)
	    unset debug(file)
	}
    }
}
#------------------------------------------------------------------------
# Background error handler
#------------------------------------------------------------------------
proc bgerror { err } {
    set trace {}
    puts_fail "$err"
    catch {set trace $::errorInfo}
    if {[string length $trace]} {
        puts_fail "*** Stack Trace ***"
        puts_fail "$trace"
    }
    TestExit
}
#------------------------------------------------------------------------
# Copy initialization
#------------------------------------------------------------------------
proc copy_init { dirname } {
    if { [ catch {
	puts_info 20 "dirname: $dirname"
	# Create temporary directory
	if { ![file exists $dirname] } {
	    file mkdir $dirname
	}
    } err ] } {
	return -code error $err
    }
}
#------------------------------------------------------------------------
# Copy deinitialization
#------------------------------------------------------------------------
proc copy_deinit { dirname } {
    if { [ catch {
	puts_info 10 "dirname: $dirname"
	if { [ file exists $dirname ] } {
	    # Remove files
	    set fail [ catch { glob $dirname/* } filelist ]
	    if {$fail} {
		TestExit
	    }
	    foreach file $filelist {
		file delete -force $file
	    }
	    # Remove directory
	    if { [file exists $dirname] } {
		file delete -force $dirname
	    }
	}
    } err ] } {
	return -code error $err
    }	
}
#------------------------------------------------------------------------
# Run md5sum for input and output files
#------------------------------------------------------------------------
proc check_md5sum { infile outfile } {
    set retval 1
    puts_info 10 "infile: $infile outfile: $outfile"
    if { ! [ file exists $infile ]  } {
	puts_fail "$infile does not exist"
	set retval 0
    }
    if { ! [ file exists $outfile ] } {
	puts_fail "$outfile does not exist"
	set retval 0
    }
    if { ! $retval } {
	return $retval
    }
    if { [ lindex [ split [exec md5sum $infile] ] 0 ] == \
	     [ lindex [ split [exec md5sum $outfile] ] 0 ] } {
	puts_pass "[clock format [clock seconds]]: MD5 checksum on $outfile"
	return 1
    } else {
	puts_fail "[clock format [clock seconds]]: MD5 checksum on $outfile"
	return 0
    }
}
#------------------------------------------------------------------------
# Implement do-while loops from Tcl mailing list (http://wiki.tcl.tk/917)
#------------------------------------------------------------------------
proc do {script arg2 {arg3 {}}} {
    #
    # Implements a "do <script> until <expression>" loop
    # The "until" keyword ist optional
    #
    # It is as fast as builtin "while" command for loops with
    # more than just a few iterations.
    #
    if {[string compare $arg3 {}]} {
        if {[string compare $arg2 until]} {
            return -code 1 "Error: do script ?until? expression"
        }
    } else {
        # copy the expression to arg3, if only
        # two arguments are supplied
        set arg3 $arg2
    }
    set ret [catch { uplevel $script } result]
    switch $ret {
        0 -
        4 {}
        3 {return}
        default {
            return -code $ret $result
        }
    }
    set ret [catch {uplevel [list while "!($arg3)" $script]} result]
    return -code $ret $result
}
#------------------------------------------------------------------------
# Error handler routine
#------------------------------------------------------------------------
proc err_handler { err } {
    set trace {}
    catch {set trace $::errorInfo}
    if { [ catch {
	switch -regexp -- $err {
	    {invalid command} {
		# Handle invalid command
		set message "\nError: invalid command.\n"
	    }
	    {contact string} {
		# Handle invalid contact string
		set message "\nError: invalid contact string.\n"
	    }
	    {canceled} {
		# Handle canceled request
		set message "\nError: canceled request.\n"
	    }
	    {no such variable} {
		set message "\nError: missing variable.\n"		
	    }
	    {not found} {
		set message "\nError: missing file.\n"
	    }
	    {[Uu]nexpected state} {
		set message "Error: Unexpected state.\n"
	    }
	    {gss_acquire_cred} {
		set message "Error: No valid proxy certificate.\n"
	    }
	    default {
		# Unknown error
		set message "Error: General: $err\n"
	    }
	}
    } err ] } {
	puts_fail "EXCEPTION: $err"
	TestExit
    }
    if {[string length $trace]} {
	append message "\n*** Stack Trace ***\n"
	append message "$trace\n"
	puts_fail $message
    }
    TestExit
}
#------------------------------------------------------------------------
# Exception handle routine
#------------------------------------------------------------------------
proc try { script1 finally script2 } {
    if { [string compare $finally {finally}] } {
	append message \
	    "syntax error: should be \"" [lindex [info level 0] 0] \
	    " script1 finally script2\""
	return -code error $message
    }
    
    set status1 [catch {
	uplevel 1 $script1
    } result1]
    
    if { $status1 == 1 } {
	set info1 $::errorInfo
	set code1 $::errorCode
    }
    
    set status2 [catch {
	uplevel 1 $script2
    } result2]
    
    switch -exact -- $status2 {
	0 {                             # TCL_OK - 'finally' was ok
	    switch -exact -- $status1 {
		0 {                     # TCL_OK - 'try' was also ok
		    return $result1
		}
		1 {                     # TCL_ERROR - 'try' failed
		    
		    return -code error \
			-errorcode $code1 \
			-errorinfo $info1 \
			$result1
		    switch -regexp -- $result1 {
		        {end of file} { 
			    puts "End of file encountered" 
			    append message \
				"end of file"
			    return -code error $message
			}
			{invalid command} {
			    # Handle invalid command
			    append message \
				"invalid command"
			    return -code error $message
			}
			{contact string} {
			    # Handle invalid contact string
			    append message \
				"invalid contact string"
			    return -code error $message
			}
			{canceled} {
			    # Handle canceled request
			    append message \
				"canceled request"
			    return -code error $message
			}
			{no such variable} {
			    append message \
				"missing variable"
			    return -code error $message
			}
			{not found} {
			    append message \
				"missing file"
			    return -code error $message
			}
			{[Uu]nexpected state} {
			    append message \
				"Unexpected state"
			    return -code error $message
			}
			default {
			    # Unknown error
			    append message \
				"Unknown error"
			    return -code error $code1
			}
		    }
		}
		2 {                     #  TCL_RETURN
		    return -code return $result1
		}
		3 {                     # TCL_BREAK
		    return -code break
		}
		4 {                     # TCL_CONTINUE
		    return -code continue
		}
		default {               # Another code
		    return -code $code $result1
		}
	    }
	}
	1 {                             # TCL_ERROR -- 'finally' failed
	    set info2 $::errorInfo
	    set code2 $::errorCode
	    append info2 "\n    (\"finally\" block)"
	    return -code error -errorcode $code2 -errorinfo $info2 \
		$result2
	}
	2 {                             # TCL_RETURN
	    # A 'return' in a 'finally' block overrides
	    # any status from the 'try' ?
	    
	    return -code return $result2
	}
	3 {                             # TCL_BREAK
	    # A 'break' in a 'finally' block overrides
	    # any status from the 'try' ?
	    
	    return -code break
	}
	4 {                             # TCL_CONTINUE
	    # A 'continue' in a 'finally' block overrides
	    # any status from the 'try' ?
	    
	    return -code break
	}
	default {                       # Another code in 'finally'
	    # Another code in a 'finally' block is returned
	    # overriding any status from the 'try'
	    
	    return -code $code $result2
	}
    }
    
}
#------------------------------------------------------------------------
# File type checking
#------------------------------------------------------------------------
# fileType --
#
#	Do some simple heuristics to determine file type.
#
#
# Arguments:
#	filename   Name of the file to test.
#
# Results:
#	type       Type of the file.  May be a list if multiple tests
#                 are positive (eg, a file could be both a directory  
#                 and a link).  In general, the list proceeds from most
#                 general (eg, binary) to most specific (eg, gif),  so
#                 the full type for a GIF file would be 
#                 "binary graphic gif"
#
#                 At present, the following types can be detected:
# 
#                       directory
#                       empty
#                       binary
#                       text
#                       script <interpreter>
#                       executable elf
#                       binary graphic [gif, jpeg, png, tiff]
#                       ps, eps, pdf
#                       html
#                       xml <doctype>
#                       message pgp
#                       bzip, gzip
#                       gravity_wave_data_frame
#                       link
#                  

proc fileType { filename } {
    ;## existence test
    if { ! [ file exists $filename ] } {
        set err "file not found: '$filename'"
        return -code error "fileType: $err"
    }
    ;## directory test
    if { [ file isdirectory $filename ] } {
        set type directory
        if { ! [ catch {file readlink $filename} ] } {
            lappend type link
        }
        return $type
    }
    ;## empty file test
    if { ! [ file size $filename ] } {
        set type empty
        if { ! [ catch {file readlink $filename} ] } {
            lappend type link
        }
        return $type
    }
    set bin_rx {[\x00-\x08\x0b\x0e-\x1f]}
    
    if { [ catch {
        set fid [ open $filename r ]
        fconfigure $fid -translation binary
        fconfigure $fid -buffersize 1024
        fconfigure $fid -buffering full
        set test [ read $fid 1024 ]
        ::close $fid 
    } err ] } {
        set err $::errorInfo
        catch { ::close $fid }
        return -code error "fileType: $err"
    }
    
    if { [ regexp $bin_rx $test ] } {
        set type binary
        set binary 1
    } else {
        set type text
        set binary 0
    }
    if { [ regexp {^\#\!\s*(\S+)} $test -> terp ] } {
        lappend type script $terp
    } elseif { $binary && [ regexp {^[\x7F]ELF} $test ] } {
        lappend type executable elf
    } elseif { $binary && [ regexp {^BZh91AY\&SY} $test ] } {
        lappend type compressed bzip
    } elseif { $binary && [ regexp {^\x1f\x8b} $test ] } {
        lappend type compressed gzip
    } elseif { $binary && [ regexp {^GIF} $test ] } {
        lappend type graphic gif
    } elseif { $binary && [ regexp {^\x89PNG} $test ] } {
        lappend type graphic png
    } elseif { $binary && [ regexp {^\xFF\xD8\xFF\xE0\x00\x10JFIF} $test ] } {
        lappend type graphic jpeg
    } elseif { $binary && [ regexp {^MM\x00\*} $test ] } {
        lappend type graphic tiff
    } elseif { $binary && [ regexp {^\%PDF\-} $test ] } {
        lappend type pdf
    } elseif { $binary && [ regexp {^ÿû\x90} $test ] } {
        lappend type music mp3
    } elseif { $binary && [ regexp {^RIFF} $test ] } {
        lappend type music wav
    } elseif { ! $binary && [ regexp -nocase {\<html\>} $test ] } {
        lappend type html       
        if { [ regexp -nocase {404\s+Not\s+Found} $test ] } {
           lappend type 404
        } elseif { [ regexp -nocase {Error\s+(\d{3})} $test -> errcode ] } {
           lappend type $errcode
        }
    } elseif { [ regexp {^\%\!PS\-} $test ] } {
       lappend type ps
       if { [ regexp { EPSF\-} $test ] } {
           lappend type eps
       }
    } elseif { [ regexp -nocase {\<\?xml} $test ] } {
        lappend type xml
        if { [ regexp -nocase {\<\!DOCTYPE\s+(\S+)} $test -> doctype ] } {
            lappend type [ string tolower $doctype ]
        }
    } elseif { [ regexp {BEGIN PGP MESSAGE} $test ] } {
        lappend type message pgp
    } elseif { $binary && [ regexp {^IGWD} $test ] } {
        lappend type gravity_wave_data_frame
    }    
    ;## lastly, is it a link?
    if { ! [ catch {file readlink $filename} ] } {
        lappend type link
    }
    return $type
}
#------------------------------------------------------------------------
#
#------------------------------------------------------------------------
proc IsVerbose { } {
    if { [ info exists ::env(TEST_VERBOSE_MODE) ] } {
	if { $::env(TEST_VERBOSE_MODE) == "true" } {
	    return 1
	} elseif { [regexp {^(\d+)$} $::env(TEST_VERBOSE_MODE)] } {
	    return $::env(TEST_VERBOSE_MODE)
	}
    }
    return -1
}
#------------------------------------------------------------------------
# Put a continuation line
#------------------------------------------------------------------------
proc puts_cont { level text } {
    if { [IsVerbose] >= $level } {
	puts "-- $text"
	flush stdout
    }
}
#------------------------------------------------------------------------
# Poceedure to dump debug message to a log file
#------------------------------------------------------------------------
proc puts_logfile { args } { 
    if { [llength $args ] == 3 } {
	set logfile [lindex $args 0]
	set mode [lindex $args 1] 
	set text [lindex $args 2]
    } 
    puts $logfile "$mode: $text"
}
#------------------------------------------------------------------------
# Poceedure to print debug message to screen
#------------------------------------------------------------------------
proc puts_debug { args } {
    if { [llength $args ] == 2 } {
	set level [lindex $args 0]
	set text [lindex $args 1]
    } else {
	set level 0
	set text $args
    }
    puts_cont $level "DEBUG: $text"
}
#------------------------------------------------------------------------
# Poceedure to print informational message to screen
#------------------------------------------------------------------------
proc puts_fail { Text } {
    puts_cont 0 "FAIL: $Text"
    set ::ExitStatus 0
}
#------------------------------------------------------------------------
# Poceedure to print failed message to screen
#------------------------------------------------------------------------
proc puts_info { args } {
    if { [llength $args ] == 2 } {
	set level [lindex $args 0]
	set text [lindex $args 1]
    } else {
	set level 0
	set text $args
    }
    puts_cont $level "INFO: $text"
}
#------------------------------------------------------------------------
# Poceedure to print pass message to screen
#------------------------------------------------------------------------
proc puts_pass { Text } {
    puts_cont 0 "PASS: $Text"
}
#------------------------------------------------------------------------
# Proceedure to terminate the test
#------------------------------------------------------------------------
proc TestExit {} {
    puts_info 10 "ExitStatus: $::ExitStatus"
    exit [expr ! $::ExitStatus]
}
#------------------------------------------------------------------------
# Common initialization code
#------------------------------------------------------------------------
# initialize the topbuilddir
#------------------------------------------------------------------------
set ::topbuilddir [ file split [pwd] ]
set ::topbuilddir [ lreplace $::topbuilddir \
			[ expr [ llength $::topbuilddir ] - 1 ] \
			[ llength $::topbuilddir ] ]
set ::ExitStatus 1
#------------------------------------------------------------------------
# Load system dependent variables
#------------------------------------------------------------------------
#------------------------------------------------------------------------
# Establish default values
#------------------------------------------------------------------------
if { [ file exists "$env(TOPSRCDIR)/common/swiglobus-test.rc"] } {
    load_rc "$env(TOPSRCDIR)/common/swiglobus-test.rc"
}
#------------------------------------------------------------------------
# Overwrite with user specified values
#------------------------------------------------------------------------
if { [ info exists env(SWIGLOBUS_TEST_CFG) ] } {
    load_rc "$env(SWIGLOBUS_TEST_CFG)"
} elseif { [ file exists "$env(HOME)/.swiglobus-test.rc"] } {
    load_rc "$env(HOME)/.swiglobus-test.rc"
}
#------------------------------------------------------------------------
# The procedure to invoke from trace 
#------------------------------------------------------------------------
proc traceProc { varName index operation } {
    upvar $varName var
    set lvl [info level]
    incr lvl -1;
    puts "Variable $varName is being modified in: '[info level $lvl]'"
    if { $lvl > 1 } {
	incr lvl -1;
	puts "Which was invoked from: '[info level $lvl]'"
    }
    puts "The current value of $varName is: $var"
}
#------------------------------------------------------------------------
# 
#------------------------------------------------------------------------
proc checkLeaks { args } {
    if { [ catch {
	set out "globals [ info globals * ]\n"
	foreach name [ namespace children ] {
	    append out "$name\n"
	    set vars [ namespace eval $name info vars ::${name}::* ]
	    foreach var $vars {
		append out "$var,value "
		if { [ namespace eval $name array exist $var ] } {
		    append out "[ namespace eval $name array get $var ]\n"
		} else {
		    append out "[ namespace eval $name set $var ]\n"
		}
	    }
	}
	set cmds [ info commands ]
	append out "checking swig name leaks in commands\n"
	foreach cmd [ info commands *_p_* ] {
	    if { [ regexp {_[a-f\d]+_p_} $cmd ] } {
		append out "$cmd\n"
	    }
	}
	set fd [ open leaks.log w ]
	puts $out
	puts $fd $out
	close $fd
    } err ] } {
	return -code error "[ myName ]: $err"
    }
}
#------------------------------------------------------------------------
# Running the test infinitely or not, default is false
#------------------------------------------------------------------------
proc IsInfinite { } {
    set SECTION general
    validate_section $SECTION [list IS_INFINITE]
    
    if { $::TestEnv($SECTION,IS_INFINITE) == "true" } {
	return 1
    } else {
	return 0
    }
    return -1
}
#------------------------------------------------------------------------
# Running the test with multiple files
#------------------------------------------------------------------------
proc IsMultiple { } {
    set SECTION general
    validate_section $SECTION [list IS_MULTIPLE]
    if { $::TestEnv($SECTION,IS_MULTIPLE) == "true" } {
	return 1
    } else {
	return 0
    }
    return -1
}
#------------------------------------------------------------------------
# Running the test with multiple files per iteration
#------------------------------------------------------------------------
proc constructFileList { } {
    # Multiple files or just a single file
    if { [ file type $::sourcedir ] != "directory" } {
	puts_info 10 "$::directory is not a directory"
    }
    if { [IsMultiple] == 1 } {
	puts_info 10 "Multiple files"
	set fail [ catch { glob $::sourcedir/* } in_filelist ]
    } else {
	puts_info 10 "Single file"
	validate_section $::SECTION [list FILENAME]
	puts_info 10 "FILENAME: $::TestEnv($::SECTION,FILENAME)"
	set fail \
	    [ catch { glob $::sourcedir/$::TestEnv($::SECTION,FILENAME) } \
		  in_filelist ]
    }
    if {$fail} {
	TestExit
    }
    return $in_filelist
}
#------------------------------------------------------------------------
# Construct output directory name based on platform and Globus flavors
#------------------------------------------------------------------------
proc constructOutputDirectory { argv } {
    # Strip off filename from tclsh input filename
    set str [split $argv /]
    set str [lindex $str [expr [llength $str]-1]]
    set str [split $str .]
    set filename [lindex $str 0]
    
    # Get OS release to determine RH9 or FC3
    if { $::tcl_platform(os) == "Linux" } {
	set fid [open /proc/sys/kernel/osrelease r]
	set osrelease [read $fid 64]
	close $fid
	if {[regexp "2\.4" $osrelease]} {
	    append filename "-"
	    append filename "rh9"
	}
	if {[regexp "2\.6" $osrelease]} {
	    append filename "-"
	    if {[regexp "FC3" $osrelease]} {
		append filename "fc3"
	    }
	    if {[regexp "FC4" $osrelease]} {
		append filename "fc4"
	    }
	}
    } elseif { $::tcl_platform(os) == "SunOS" } {
	append filename "-"
	append filename $::tcl_platform(os)
	append filename $::tcl_platform(osVersion)
    } else {
	append filename "-"
	append filename "unknown"
    }

    # Get GT version, GT3.2.1 or GT4.0.0
    if {[string length [auto_execok globus-version]] != 0} {
	set gt_version [exec globus-version]
	if {[regexp "4\.0\.1" $gt_version]} {
	    append filename "-"
	    append filename "gt4"
	}
    } else {
	append filename "-"
	append filename "gt3"
    }
    
    # Get GT information for threaded or non-threaded version
    set config [file join [file dirname [file dirname [pwd]]] config.h]
    if {[file exists $config]} {
	set fid [open $config r]
	set str [read $fid 5120]
	close $fid
	set found [regexp -line "(\[^\n]*HAVE_GLOBUS_THREADED\[^\n]*)" \
		       $str full]
	if { $found } {
	    set threaded [regexp {[0-9]} $full]
	    if { $threaded } {
		append filename "-"
		append filename "threaded"
	    } else {
		append filename "-"
		append filename "nonthreaded"
	    }
	}
    }
    return $filename
}
#------------------------------------------------------------------------
# To verify proxy cert
#------------------------------------------------------------------------
proc verifyProxy {} {
    set prog grid-proxy-info
    set progpath [ auto_execok $prog ]
    if { ! [string length $progpath] } {
	puts_fail "$prog not found: unable to verify"
	TestExit
    }
    
    set rc [ catch { exec grid-proxy-info -timeleft } data ]

    set data [split $data \n] 
    set timeleft [lindex $data 0]

    puts_info 10 "timeleft: $timeleft"
    
    if { ![string length $timeleft] } {
	puts_fail "You have no valid proxy certificate"
	TestExit
    }
}

#------------------------------------------------------------------------
# To verify remote server
#------------------------------------------------------------------------
proc verifyServer { host port } {
    if { [ catch {
	set sock [ socket $host $port ]
	close $sock
    } err ] } {
	puts_fail "EXCEPTION: $err"
	TestExit
    }
}
