#!/bin/sh
# ALL the continued lines following this one are interpreted
# by the bourne shell and ignored by Tcl, which picks up
# execution after the exec line.\
exec tclsh "$0" ${1+"$@"}

#------------------------------------------------------------------------
# Initialize the test
#------------------------------------------------------------------------
source "$env(SRCDIR)/test_utils.tcl"

foreach lib { globus_module \
		  globus_error \
		  globus_object \
		  globus_xio } {
    load [ eval file join $topbuilddir "swig/.libs/libtcl${lib}.so" ]
}
#------------------------------------------------------------------------
# Sanity check to verify that the version of Globus being used supports
#   the xio library.
#------------------------------------------------------------------------
if { [ info exists globus_i_xio_module ] == 0 } {
    #--------------------------------------------------------------------
    # No xio extensions as is the case with globus 2.4.x
    #--------------------------------------------------------------------
    TestExit
}

#------------------------------------------------------------------------
# Initialize necessary variables 
#------------------------------------------------------------------------
set SECTION xio_test
validate_section $SECTION [list FILENAME BUFFERSIZE WAITFORNBYTES]

set bufsize $TestEnv($SECTION,BUFFERSIZE)
puts_info 10 "bufsize: $bufsize"
set waitforbytes $TestEnv($SECTION,WAITFORNBYTES)
puts_info 10 "waitforbytes: $waitforbytes"
set filename $TestEnv($SECTION,FILENAME)
puts_info 10 "filename: $filename"
set contact_string [ lindex $argv 0 ]
puts_info 10 "contact_string: $contact_string"

if { [ string length $contact_string ] == 0 } {
    set mode -server
} else {
    set mode -client
}
puts_info 10 "mode: $mode"
#------------------------------------------------------------------------
# Client proc
#------------------------------------------------------------------------
namespace eval client {
    proc client_operation { } {
	if { [ catch {

	    puts_info 10 "client pid: [pid]"
	    
	    # Stack construction
	    lassign [ globus_xio_stack_init NULL ] status filestack
	    lassign [ globus_xio_stack_init NULL ] status udpstack
	    
	    # Push drivers onto stack
	    lassign [ globus_xio_stack_push_driver \
			  $filestack \
			  $::filedriver \
			 ] status
	    
	    lassign [ globus_xio_stack_push_driver \
			  $udpstack \
			  $::udpdriver \
			 ] status
	    
	    # File handle construction
	    lassign [ globus_xio_handle_create $filestack ] status filehandle
	    
	    # Network handle construction
	    lassign [ globus_xio_handle_create $udpstack ] status udphandle
	    
	    # UDP attributes
	    lassign [ globus_xio_attr_init ] result udpattr

	    if { 1 } {
		puts_info 10 "globus_xio_attr_cntl: GLOBUS_XIO_UDP_SET_NO_IPV6"
		lassign [ globus_xio_attr_cntl \
			      $udpattr \
			      $::udpdriver \
			      $::GLOBUS_XIO_UDP_SET_NO_IPV6 \
			      $::GLOBUS_TRUE \
			     ] status
		puts_info 10 "status: $status"
	    }
	    # Network handle
	    lassign [ globus_xio_handle_create $udpstack ] status udphandle
	    
	    # File handle
	    lassign [ globus_xio_handle_create $filestack ] status filehandle
	    
	    # Open file handle
	    lassign [ globus_xio_open $filehandle $::read_cs NULL ] status
	    puts_info 10 "globus_xio_open on filehandle returns: $status"
	    
	    lassign [ globus_xio_open \
			  $udphandle \
			  $::contact_string \
			  $udpattr \
			 ] status
	    
	    # Blocking read and write
	    set orig_filesize [ file size $::read_cs ]
	    set remainingbytestoread [ expr $orig_filesize ]
	    set totalbufsizes [ expr $::bufsize ]
	    
	    do {
		# Read from the file
		lassign [ globus_xio_read \
			      $filehandle \
			      $::bufsize \
			      $::waitforbytes \
			      NULL \
			     ] read_status buffer
		puts_info 30 "client buffer length = [string length $buffer]"
		
		# Write to network
		lassign [ globus_xio_write \
			      $udphandle \
			      $buffer \
			      $::waitforbytes \
			      NULL \
			     ] write_status nwritten
		
		set remainingbytestoread [ expr { $remainingbytestoread - \
						      $totalbufsizes } ]
	    } until { $remainingbytestoread < 0 }

	    set ::done 1
	    
	    # Destroy attributes
	    globus_xio_attr_destroy $udpattr
	    
	    # Close both handles 
	    globus_xio_close $udphandle NULL
	    globus_xio_close $filehandle NULL
	    
	    # Destroy both stacks
	    globus_xio_stack_destroy $filestack
	    globus_xio_stack_destroy $udpstack
	} err ] } {
	    return -code error $err
	}
    }
}
#------------------------------------------------------------------------
# Server proc
#------------------------------------------------------------------------
namespace eval server {
    variable filestack
    variable udpstack
    variable udpattr
    variable filehandle
    variable udphandle
    variable dd

    proc init_server { } {
	variable filestack
	variable udpstack
	variable udpattr
	variable filehandle
	variable udphandle
	variable dd

	# Stack construction
	puts_info 10 "Server stack construction"
	lassign [ globus_xio_stack_init NULL ] status udpstack
	lassign [ globus_xio_stack_init NULL ] status filestack
	
	# Push drivers onto stack
	puts_info 10 "Push drivers onto server stack"
	lassign [ globus_xio_stack_push_driver \
		      $filestack \
		      $::filedriver \
		     ] status
	
	lassign [ globus_xio_stack_push_driver \
		      $udpstack \
		      $::udpdriver \
		     ] status
	
	# UDP attributes
	puts_info 10 "Set UDP attributes"
	lassign [ globus_xio_attr_init ] status udpattr

	puts_info 10 "globus_xio_attr_cntl: GLOBUS_XIO_UDP_SET_NO_IPV6"
	lassign [ globus_xio_attr_cntl \
		      $udpattr \
		      $::udpdriver \
		      $::GLOBUS_XIO_UDP_SET_NO_IPV6 \
		      $::GLOBUS_TRUE \
		     ] status
	puts_info 10 "status: $status"
	
	if { 0 } {
	    lassign [ globus_xio_attr_cntl \
			  $udpattr \
			  $::udpdriver \
			  $::GLOBUS_XIO_UDP_SET_PORT \
			  50000 \
			 ] status
	    puts_info 10 "globus_xio_attr_cntl: GLOBUS_XIO_UDP_SET_PORT"
	    puts_info 10 "status: $status"
	    
	    lassign [ globus_xio_attr_cntl \
			  $udpattr \
			  $::udpdriver \
			  $::GLOBUS_XIO_UDP_GET_PORT \
			 ] status listener_port
	    puts_info 10 "globus_xio_attr_cntl: GLOBUS_XIO_UDP_GET_PORT"
	    puts_info 10 "status: $status"
	    puts_info 10 "listener_port: $listener_port"
	}
	
	if { 0 } {
	    # Set UDP Send Buffer
	    set udp_sendbuffer 512
	    
	    globus_xio_attr_cntl \
		$udpattr \
		$::udpdriver \
		$::GLOBUS_XIO_UDP_SET_SNDBUF \
		$udp_sendbuffer 
	    
	    # Set UDP Receive Buffer
	    set udp_receivebuffer 512
	    
	    globus_xio_attr_cntl \
		$udpattr \
		$::udpdriver \
		$::GLOBUS_XIO_UDP_SET_RCVBUF \
		$udp_receivebuffer
	}
	
	# File handle construction
	lassign [ globus_xio_handle_create $filestack ] status filehandle
	
	# Network handle construction
	lassign [ globus_xio_handle_create $udpstack ] status udphandle
	
	# Open UDP handle
	lassign [ globus_xio_open $udphandle NULL $udpattr ] status
	
	#lassign [ globus_xio_data_descriptor_init $udphandle ] result dd
	
	# Open file handle
	lassign [ globus_xio_open $filehandle $::write_cs NULL ] status

	lassign [ globus_xio_handle_cntl \
		      $udphandle \
		      $::udpdriver \
		      $::GLOBUS_XIO_UDP_GET_NUMERIC_CONTACT \
		     ] status contact_string
	
	# Output server contact string info
	return $contact_string
    }


    proc server_data_operation { } {
	variable filestack
	variable udpstack
	variable udpattr
	variable filehandle
	variable udphandle
	variable dd

	if { [ catch {
	    # Blocking read and write
	    puts_info 10 "Blocking read and write"
	    
	    set bytes_written 0

	    do {
		    
		# Read from the network
		lassign [ globus_xio_read \
			      $udphandle \
			      $::bufsize \
			      $::waitforbytes \
			      NULL \
			     ] read_status buffer
		
		puts_info 30 "server buffer length: [string length $buffer]"

		# Write to the file
		lassign [ globus_xio_write \
			      $filehandle \
			      $buffer \
			      $::waitforbytes \
			      NULL \
			     ] write_status bytes_written
		puts_info 30 "bytes_written: $bytes_written"
	    } until { $bytes_written < $::bufsize } 
	    
	    # Destroy data descriptor attr
	    #globus_xio_data_descriptor_destroy $dd
	    
	    # Close UDP handle
	    globus_xio_close $udphandle $udpattr
	    
	    # Close file handle
	    globus_xio_close $filehandle NULL
	    
	    # Destroy stacks
	    globus_xio_stack_destroy $filestack 
	    globus_xio_stack_destroy $udpstack
	    
	    # Destroy attributes
	    globus_xio_attr_destroy $udpattr

	    puts_info 10 "Server is completed normally"
	} err ] } {
	    puts_info 10 "Server is completed abnormally"
	    return -code error $err
	}
    }
}
#------------------------------------------------------------------------
# Actual demo
#------------------------------------------------------------------------
if { [ catch {
    puts_info 10 "[info nameofexecutable]"
    
    # Set up local directory name
    set localdir [file join /tmp \
		      [file join "$env(USER)" \
			   [constructOutputDirectory $argv0]]]
    puts_info 10 "localdir: $localdir"
    
    set write_cs [ file join $localdir $filename ] 
    puts_info 10 "write_cs: $write_cs"
    
    set read_cs [ file join [ file dirname [ pwd ] ] swig $filename ]
    puts_info 10 "read_cs: $read_cs"

    # Create temporary directory
    copy_init $localdir

    # Activate Globus XIO module
    puts_info 10 "Activate Globus XIO module"
    globus_module_activate $globus_i_xio_module

    # Load transport and transform driver
    puts_info 10 "Load transport and transform driver"
    lassign [ globus_xio_driver_load udp ] result udpdriver
    lassign [ globus_xio_driver_load file ] result filedriver

    set done 0
    
    # Runs server process
    switch -- $mode {
	-client {
	    client::client_operation
	}
	
	-server { 
	    puts_info 10 "argc: $argc"
	    set cs [ server::init_server ]
	    puts_info 10 "cs: $cs"
	
	    # Exec client process
	    catch { exec $argv0 $cs &  } client_pid

	    puts_info 10 "client_pid: $client_pid"

	    # Continue with the server
	    server::server_data_operation

	    puts_info 10 "Server is completed"

	    # Run MD5 checksum before exiting
	    check_md5sum $read_cs $write_cs

	    # Deinitialization step
	    copy_deinit $localdir
	}
    }

    vwait done
    
    # Unload both drivers
    globus_xio_driver_unload $filedriver 
    globus_xio_driver_unload $udpdriver

    # Deactivate Globus XIO module
    globus_module_deactivate $globus_i_xio_module
} err ] } {
    err_handler $err
}
#------------------------------------------------------------------------
# Exit with 0 if no tests failed, 1 otherwise
#------------------------------------------------------------------------
TestExit
